# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


from unittest import mock

import pika
from twisted.internet import defer

from fedora_messaging.twisted.protocol import FedoraMessagingProtocolV2


class MockChannel(mock.Mock):
    """A mock object with Channel-specific methods that return Deferreds."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        deferred_methods = (
            "basic_qos",
            "confirm_delivery",
            "exchange_declare",
            "queue_bind",
            "basic_ack",
            "basic_nack",
            "basic_publish",
            "basic_cancel",
            "close",
        )
        for method in deferred_methods:
            setattr(
                self,
                method,
                mock.Mock(side_effect=lambda *a, **kw: defer.succeed(None)),
            )
        self.queue_declare = mock.Mock(
            side_effect=lambda **kw: defer.succeed(
                pika.frame.Method(0, pika.spec.Queue.DeclareOk(queue=kw["queue"]))
            )
        )
        # self.queue_object = mock.Mock(name="queue_object")
        self.queue_object = defer.DeferredQueue()
        self.basic_consume = mock.Mock(
            side_effect=lambda **kw: defer.succeed((self.queue_object, "consumer-tag"))
        )


class MockProtocol(FedoraMessagingProtocolV2):
    """A Protocol object that mocks the underlying channel and impl."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._impl = mock.Mock(name="_impl")
        self._impl.is_closed = True
        self._channel = MockChannel(name="_channel")
        self.channel = mock.Mock(name="channel", side_effect=lambda: defer.succeed(self._channel))

    def _register_consumer(self, consumer):
        consumer._protocol = self
        consumer._channel = self._channel
        self._consumers[consumer.queue] = consumer



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
The ``fedora-messaging`` `Click`_ CLI.

.. _Click: http://click.pocoo.org/
"""

import errno
import importlib
import logging
import logging.config
import os
import sys

import click
import requests
from twisted.internet import asyncioreactor, error

from fedora_messaging import message


try:
    asyncioreactor.install()
except error.ReactorAlreadyInstalledError:
    # The tests install a reactor before importing this module
    from twisted.internet import reactor

    if not isinstance(reactor, asyncioreactor.AsyncioSelectorReactor):
        raise

from twisted.internet import reactor
from twisted.python import log as legacy_twisted_log

from . import api, config, exceptions
from .message import dumps, loads


_log = logging.getLogger(__name__)

_conf_help = (
    "Path to a valid configuration file to use in place of the "
    "configuration in /etc/fedora-messaging/config.toml."
)
_app_name_help = (
    "The name of the application, used by the AMQP client to identify itself to "
    "the broker. This is purely for administrator convenience to determine what "
    "applications are connected and own particular resources."
)
_callback_help = (
    "The Python path to the callable object to execute when a message arrives. "
    "The Python path should be in the format ``module.path:object_in_module`` "
    "and should point to either a function or a class. Consult the API "
    "documentation for the interface required for these objects."
)
_callback_file_help = (
    "The path of a Python file that contains the callable object to"
    " execute when the message arrives. This should be in the format "
    '"my/script.py:object_in_file". Providing this overrides the callback set '
    'by the "--callback" argument or configuration file. It executes the entire '
    "Python file to load the callable object."
)
_routing_key_help = (
    "The AMQP routing key to use with the queue. This controls what messages are "
    "delivered to the consumer. Can be specified multiple times; any message "
    "that matches at least one will be placed in the message queue. "
)
_queue_name_help = (
    "The name of the message queue in AMQP. Can contain ASCII letters, digits, "
    "hyphen, underscore, period, or colon. If one is not specified, a unique "
    "name will be created for you."
)
_exchange_help = (
    "The name of the exchange to bind the queue to. Can contain ASCII letters, "
    "digits, hyphen, underscore, period, or colon."
)
_publish_exchange_help = (
    "The name of the exchange to publish to. Can contain ASCII letters, "
    "digits, hyphen, underscore, period, or colon."
)
_datagrepper_help = (
    "The URL of the datagreeper instance to use, " "defaults to the production environment"
)
_limit_help = "The maximum number of messages to record."


# Global variable used to set the exit code in error handlers, then let
# the reactor shut down gracefully, then exit with the proper code.
_exit_code = 0


@click.group()
@click.option("--conf", envvar="FEDORA_MESSAGING_CONF", help=_conf_help)
def cli(conf):
    """The fedora-messaging command line interface."""
    if conf:
        if not os.path.isfile(conf):
            raise click.exceptions.BadParameter(f"{conf} is not a file")
        try:
            config.conf.load_config(config_path=conf)
        except exceptions.ConfigurationException as e:
            raise click.exceptions.BadParameter(str(e)) from e
    twisted_observer = legacy_twisted_log.PythonLoggingObserver()
    twisted_observer.start()
    config.conf.setup_logging()


@cli.command()
@click.option("--app-name", help=_app_name_help)
@click.option("--callback-file", help=_callback_file_help)
@click.option("--callback", help=_callback_help)
@click.option("--routing-key", help=_routing_key_help, multiple=True)
@click.option("--queue-name", help=_queue_name_help)
@click.option("--exchange", help=_exchange_help)
def consume(exchange, queue_name, routing_key, callback, callback_file, app_name):
    """Consume messages from an AMQP queue using a Python callback."""
    # The configuration validates these are not null and contain all required keys
    # when it is loaded.
    if callback_file:
        callback = _callback_from_filesystem(callback_file)
    else:
        callback = _callback_from_python_path(callback)
    _consume(exchange, queue_name, routing_key, callback, app_name)


def _consume(exchange, queue_name, routing_key, callback, app_name):
    """
    The actual consume code, which expects an actual callable object.
    This lets various consume-based commands share the setup code. Anything
    that accepts None loads the defaults from the configuration.

    Args:
        exchange (str): The AMQP message exchange to bind to, or None.
        queue_name (str): The queue name to use, or None.
        routing_key (str): The routing key to use, or None.
        callback (callable): A callable object to use for the callback.
        app_name (str): The application name to use, or None.
    """
    bindings = config.conf["bindings"]
    queues = config.conf["queues"]

    if queue_name:
        queues = {queue_name: config.conf["queues"][config._default_queue_name]}
        for binding in bindings:
            binding["queue"] = queue_name

    if exchange:
        for binding in bindings:
            binding["exchange"] = exchange

    if routing_key:
        for binding in bindings:
            binding["routing_keys"] = routing_key

    if app_name:
        config.conf["client_properties"]["app"] = app_name

    try:
        deferred_consumers = api.twisted_consume(callback, bindings=bindings, queues=queues)
        deferred_consumers.addCallback(_consume_callback)
        deferred_consumers.addErrback(_consume_errback)
    except ValueError as e:
        raise click.exceptions.BadOptionUsage("callback", str(e)) from e

    reactor.run()
    sys.exit(_exit_code)


def _callback_from_filesystem(callback_file):
    """
    Load a callable from a Python script on the file system.

    Args:
        callback_file (str): The callback as a filesystem path and callable name
            separated by a ":". For example, "my/python/file.py:printer".

    Raises:
        click.ClickException: If the object cannot be loaded.

    Returns:
        callable: The callable object.
    """
    try:
        file_path, callable_name = callback_file.strip().split(":")
    except ValueError as e:
        raise click.ClickException(
            "Unable to parse the '--callback-file' option; the "
            'expected format is "path/to/file.py:callable_object" where '
            '"callable_object" is the name of the function or class in the '
            "Python file"
        ) from e

    try:
        file_namespace = {}
        with open(file_path, "rb") as fd:
            try:
                # Using "exec" is generally a Bad Idea (TM), so bandit is upset at
                # us. In this case, it seems like a Good Idea (TM), but I might be
                # wrong. Sorry.
                exec(compile(fd.read(), file_path, "exec"), file_namespace)  # noqa: S102
            except Exception as e:
                raise click.ClickException(
                    f"The {file_path} file raised the following exception during execution: {e}"
                ) from e

        if callable_name not in file_namespace:
            err = f"The '{callable_name}' object was not found in the '{file_path}' file."
            raise click.ClickException(err)
        else:
            return file_namespace[callable_name]
    except OSError as e:
        raise click.ClickException(f"An IO error occurred: {e}") from e


def _callback_from_python_path(callback):
    """
    Load a callable from a Python path.

    Args:
        callback_file (str): The callback as a Python path and callable name
            separated by a ":". For example, "my_package.my_module:printer".

    Raises:
        click.ClickException: If the object cannot be loaded.

    Returns:
        callable: The callable object.
    """
    callback_path = callback or config.conf["callback"]
    if not callback_path:
        raise click.ClickException(
            "A Python path to a callable object that accepts the message must be provided"
            ' with the "--callback" command line option or in the configuration file'
        )
    try:
        module, cls = callback_path.strip().split(":")
    except ValueError as e:
        raise click.ClickException(
            f"Unable to parse the callback path ({callback_path}); the "
            'expected format is "my_package.module:'
            'callable_object"'
        ) from e
    try:
        module = importlib.import_module(module)
    except ImportError as e:
        provider = "--callback argument" if callback else "configuration file"
        raise click.ClickException(
            f"Failed to import the callback module ({e}) provided in the {provider}"
        ) from e

    try:
        callback_object = getattr(module, cls)
    except AttributeError as e:
        raise click.ClickException(
            f"Unable to import {callback_path} ({e}); is the package installed? The python "
            'path should be in the format "my_package.module:callable_object"'
        ) from e
    _log.info("Starting consumer with %s callback", callback_path)
    return callback_object


def _consume_errback(failure):
    """Handle any errors that occur during consumer registration."""
    global _exit_code
    if failure.check(exceptions.BadDeclaration):
        _log.error(
            "Unable to declare the %s object on the AMQP broker. The "
            "broker responded with %s. Check permissions for your user.",
            failure.value.obj_type,
            failure.value.reason,
        )
        _exit_code = 10
    elif failure.check(exceptions.PermissionException):
        _exit_code = 15
        _log.error(
            "The consumer could not proceed because of a permissions problem: %s",
            str(failure.value),
        )
    elif failure.check(exceptions.ConnectionException):
        _exit_code = 14
        _log.error(failure.value.reason)
    else:
        _exit_code = 11
        _log.exception(
            "An unexpected error (%r) occurred while registering the "
            "consumer, please report this bug.",
            failure.value,
        )
    try:
        reactor.stop()
    except error.ReactorNotRunning:
        pass


def _consume_callback(consumers):
    """
    Callback when consumers are successfully registered.

    This simply registers callbacks for consumer.result deferred object which
    fires when the consumer stops.

    Args
        consumers (list of fedora_messaging.api.Consumer):
            The list of consumers that were successfully created.
    """
    for consumer in consumers:

        def errback(failure, consumer=consumer):
            global _exit_code
            if failure.check(exceptions.HaltConsumer):
                _exit_code = failure.value.exit_code
                if _exit_code:
                    _log.error(
                        "Consumer halted with non-zero exit code (%d): %s",
                        _exit_code,
                        str(failure.value.reason),
                    )
            elif failure.check(exceptions.ConsumerCanceled):
                _exit_code = 12
                _log.error(
                    "The consumer was canceled server-side, check with system administrators."
                )
            elif failure.check(exceptions.PermissionException):
                _exit_code = 15
                _log.error(
                    "The consumer could not proceed because of a permissions problem: %s",
                    str(failure.value),
                )
            else:
                _exit_code = 13
                _log.error("Unexpected error occurred in consumer %r: %r", consumer, failure)
            try:
                reactor.stop()
            except error.ReactorNotRunning:
                pass

        def callback(consumer):
            _log.info("The %r consumer halted.", consumer)
            if all([c.result.called for c in consumers]):
                _log.info("All consumers have stopped; shutting down.")
                try:
                    # Last consumer out shuts off the lights
                    reactor.stop()
                except error.ReactorNotRunning:
                    pass

        consumer.result.addCallbacks(callback, errback)


@cli.command()
@click.option("--exchange", help=_publish_exchange_help)
@click.argument("file", type=click.File("r"))
def publish(exchange, file):
    """Publish messages to an AMQP exchange from a file."""
    for msgs_json_str in file:
        try:
            messages = loads(msgs_json_str)
        except exceptions.ValidationError as e:
            raise click.BadArgumentUsage(f"Unable to validate message: {e}") from e

        for msg in messages:
            click.echo(f"Publishing message with topic {msg.topic}")
            try:
                api.publish(msg, exchange)
            except (exceptions.PublishReturned, exceptions.PublishForbidden) as e:
                click.echo(f"Unable to publish message: {e}")
                sys.exit(errno.EREMOTEIO)
            except exceptions.PublishTimeout as e:
                click.echo(f"Unable to connect to the message broker: {e}")
                sys.exit(errno.ECONNREFUSED)
            except exceptions.PublishException as e:
                click.echo(f"A general publish exception occurred: {e}")
                sys.exit(1)


class Recorder:
    """
    A simple callback class that records messages.

    Attributes:
        counter (int): The number of messages this callback has recorded.
        messages (list): The list of messages received.

    Args:
        limit (int): The maximum number of messages to record.
        file (file): The file object with write rights.
    """

    def __init__(self, limit, file):
        self.counter = 0
        self._limit = limit
        self._file = file
        if limit:
            self._bar = click.progressbar(length=limit)

    def collect_message(self, message):
        """
        Collect received messages.

        Args:
            message (message.Message): The received message.

        Raises:
            fedora_messaging.exceptions.HaltConsumer: Raised if the number of received
                messages reach the limit or if collected messages serialization is impossible.
        """
        try:
            json_str = dumps(message)
        except exceptions.ValidationError as e:
            click.echo(f"Unable to save messages to file: {e}")
            raise exceptions.HaltConsumer(exit_code=1, requeue=False) from e
        else:
            self._file.write(json_str)

        self.counter += 1
        if self._limit:
            self._bar.update(1)
            if self._limit <= self.counter:
                raise exceptions.HaltConsumer(exit_code=0, requeue=False)


@cli.command()
@click.argument("file", type=click.File("w"))
@click.option("--limit", help=_limit_help, type=click.IntRange(1))
@click.option("--app-name", help=_app_name_help, default="recorder")
@click.option("--routing-key", help=_routing_key_help, multiple=True)
@click.option("--queue-name", help=_queue_name_help)
@click.option("--exchange", help=_exchange_help)
def record(exchange, queue_name, routing_key, app_name, limit, file):
    """Record messages from an AMQP queue to provided file."""
    messages_recorder = Recorder(limit, file)
    _consume(exchange, queue_name, routing_key, messages_recorder.collect_message, app_name)


DEFAULT_DATAGREPPER_URL = "https://apps.fedoraproject.org/datagrepper"


@cli.command()
@click.argument("message_id")
@click.option(
    "--datagrepper-url",
    help=_datagrepper_help,
    default=DEFAULT_DATAGREPPER_URL,
    show_default=True,
)
def replay(message_id, datagrepper_url):
    """Replay a message from Datagrepper by its message ID"""
    try:
        message_data = _get_message(message_id, datagrepper_url)
    except requests.HTTPError as e:
        raise click.ClickException(f"Failed to retrieve message from Datagrepper: {e}") from e

    if message_data:
        # Disable the topic prefix, the loaded message already has everything.
        config.conf["topic_prefix"] = ""
        api.publish(message.load_message(message_data))
        click.echo(f"Message with ID {message_id} has been successfully replayed.")


def _get_message(message_id, datagrepper_url):
    """Fetch a message by ID from Datagreeper"""
    url = f"{datagrepper_url}/id?id={message_id}&is_raw=true"
    response = requests.get(url, timeout=5)
    response.raise_for_status()
    return response.json()



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Twisted Service to start and stop the Fedora Messaging Twisted Factory.

This Service makes it easier to build a Twisted application that embeds a
Fedora Messaging component. See the ``verify_missing`` service in
fedmsg-migration-tools for a use case.

See https://twistedmatrix.com/documents/current/core/howto/application.html
"""


import logging
import ssl

import pika
from pika import SSLOptions
from twisted.application import service
from twisted.application.internet import SSLClient, TCPClient
from twisted.internet import defer
from twisted.internet import ssl as twisted_ssl

from .. import config, exceptions
from .factory import FedoraMessagingFactoryV2


_log = logging.getLogger(__name__)


class FedoraMessagingServiceV2(service.MultiService):
    """
    A Twisted service to connect to the Fedora Messaging broker.

    Args:
        amqp_url (str): URL to use for the AMQP server.
        publish_confirms (bool): If true, use the RabbitMQ publisher confirms
            AMQP extension.
    """

    name = "fedora-messaging-servicev2"

    def __init__(self, amqp_url=None, publish_confirms=True):
        """Initialize the service."""
        service.MultiService.__init__(self)
        self._parameters = pika.URLParameters(amqp_url or config.conf["amqp_url"])
        self._confirms = publish_confirms
        if amqp_url.startswith("amqps"):
            _configure_tls_parameters(self._parameters)
        if self._parameters.client_properties is None:
            self._parameters.client_properties = config.conf["client_properties"]

        factory = FedoraMessagingFactoryV2(self._parameters, self._confirms)
        if self._parameters.ssl_options:
            self._service = SSLClient(
                host=self._parameters.host,
                port=self._parameters.port,
                factory=factory,
                contextFactory=_ssl_context_factory(self._parameters),
            )
        else:
            self._service = TCPClient(
                host=self._parameters.host, port=self._parameters.port, factory=factory
            )
        self._service.factory = factory
        name = "{}{}:{}".format(
            "ssl:" if self._parameters.ssl_options else "",
            self._parameters.host,
            self._parameters.port,
        )
        self._service.setName(name)
        self._service.setServiceParent(self)

    @defer.inlineCallbacks
    def stopService(self):
        """
        Gracefully stop the service.

        Returns:
            defer.Deferred: a Deferred which is triggered when the service has
                finished shutting down.
        """
        self._service.factory.stopTrying()
        yield self._service.factory.stopFactory()
        yield service.MultiService.stopService(self)


def _configure_tls_parameters(parameters):
    """
    Configure the pika connection parameters for TLS based on the configuration.

    This modifies the object provided to it. This accounts for whether or not
    the new API based on the standard library's SSLContext is available for
    pika.

    Args:
        parameters (pika.ConnectionParameters): The connection parameters to apply
            TLS connection settings to.
    """
    cert = config.conf["tls"]["certfile"]
    key = config.conf["tls"]["keyfile"]
    if cert and key:
        _log.info(
            "Authenticating with server using x509 (certfile: %s, keyfile: %s)",
            cert,
            key,
        )
        parameters.credentials = pika.credentials.ExternalCredentials()
    else:
        cert, key = None, None

    ssl_context = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH)
    if config.conf["tls"]["ca_cert"]:
        try:
            ssl_context.load_verify_locations(cafile=config.conf["tls"]["ca_cert"])
        except ssl.SSLError as e:
            raise exceptions.ConfigurationException(
                f'The "ca_cert" setting in the "tls" section is invalid ({e})'
            ) from e
    ssl_context.minimum_version = ssl.TLSVersion.TLSv1_2
    ssl_context.check_hostname = True
    if cert and key:
        try:
            ssl_context.load_cert_chain(cert, key)
        except ssl.SSLError as e:
            raise exceptions.ConfigurationException(
                f'The "keyfile" setting in the "tls" section is invalid ({e})'
            ) from e
    parameters.ssl_options = SSLOptions(ssl_context, server_hostname=parameters.host)


def _ssl_context_factory(parameters):
    """
    Produce a Twisted SSL context object from a pika connection parameter object.
    This is necessary as Twisted manages the connection, not Pika.

    Args:
        parameters (pika.ConnectionParameters): The connection parameters built
            from the fedora_messaging configuration.
    """
    client_cert = None
    ca_cert = None
    key = config.conf["tls"]["keyfile"]
    cert = config.conf["tls"]["certfile"]
    ca_file = config.conf["tls"]["ca_cert"]
    if ca_file:
        with open(ca_file, "rb") as fd:
            # Open it in binary mode since otherwise Twisted will immediately
            # re-encode it as ASCII, which won't work if the cert bundle has
            # comments that can't be encoded with ASCII.
            ca_cert = twisted_ssl.Certificate.loadPEM(fd.read())
    if key and cert:
        # Note that _configure_tls_parameters sets the auth mode to EXTERNAL
        # if both key and cert are defined, so we don't need to do that here.
        with open(key) as fd:
            client_keypair = fd.read()
        with open(cert) as fd:
            client_keypair += fd.read()
        client_cert = twisted_ssl.PrivateCertificate.loadPEM(client_keypair)

    hostname = parameters.host
    try:
        context_factory = twisted_ssl.optionsForClientTLS(
            hostname,
            trustRoot=ca_cert or twisted_ssl.platformTrust(),
            clientCertificate=client_cert,
            extraCertificateOptions={"raiseMinimumTo": twisted_ssl.TLSVersion.TLSv1_2},
        )
    except AttributeError:
        # Twisted 12.2 path for EL7 :(
        context_factory = twisted_ssl.CertificateOptions(
            certificate=client_cert.original,
            privateKey=client_cert.privateKey.original,
            caCerts=[ca_cert.original] or twisted_ssl.platformTrust(),
            verify=True,
            requireCertificate=True,
            verifyOnce=False,
            enableSessions=False,
        )

    return context_factory



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


from unittest import mock

import pika
import pytest
from twisted.internet import defer
from twisted.internet.error import ConnectionDone, ConnectionLost

from fedora_messaging.exceptions import ConnectionException
from fedora_messaging.twisted.consumer import Consumer
from fedora_messaging.twisted.factory import ConsumerRecord, FedoraMessagingFactoryV2


try:
    import pytest_twisted  # noqa: F401
except ImportError:
    pytest.skip("pytest-twisted is missing, skipping tests", allow_module_level=True)


class TestFactoryV2:
    def setup_method(self, method):
        self.protocol = mock.Mock()
        self.protocol.ready = defer.Deferred()
        self.protocol.is_closed = False
        protocol_class = mock.Mock(side_effect=lambda *a, **kw: self.protocol)
        self.factory = FedoraMessagingFactoryV2(
            mock.Mock(name="parameters"), {"binding key": "binding value"}
        )
        self.factory.protocol = protocol_class

    def test_buildProtocol(self):
        """Assert buildProtocol associates the factory"""
        protocol = self.factory.buildProtocol(None)
        assert protocol.factory == self.factory

    def test_when_connected(self):
        """Assert when_connected returns the current client once _client_deferred fires"""
        self.factory.buildProtocol(None)
        d = self.factory.when_connected()

        def assert_equal(a, b):
            assert a == b

        def assert_is(a, b):
            assert a is b

        d.addCallback(lambda client: assert_equal(self.factory._client, client))
        d.addCallback(lambda _: assert_is(self.factory._client_deferred.called, True))
        self.protocol.ready.callback(None)
        return d

    def test_buildProtocol_twice(self):
        """Assert buildProtocol works when reconnecting"""

        def _get_protocol(*a, **kw):
            protocol = mock.Mock(name="protocol mock")
            protocol.ready = defer.succeed(None)
            return protocol

        self.factory.protocol = _get_protocol
        connector = mock.Mock()
        connected_d = self.factory.when_connected()
        self.factory.buildProtocol(None)
        self.factory.clientConnectionLost(connector, None)
        with mock.patch("fedora_messaging.twisted.factory._std_log") as mock_log:
            protocol = self.factory.buildProtocol(None)
        assert not mock_log.exception.called
        assert not mock_log.error.called
        d = defer.DeferredList([connected_d, protocol.ready], fireOnOneErrback=True)
        d.addErrback(lambda f: f.value.subFailure)
        return d

    def _test_when_connected_disconnected(self, error_class, error_msg):
        """Assert when_connected errbacks on disconnections."""

        def _get_protocol(*a, **kw):
            protocol = mock.Mock(name="protocol mock")
            # Disconnect immediately
            protocol.ready = defer.fail(error_class())
            return protocol

        def _check(f):
            f.trap(ConnectionException)
            # Make sure a new Deferred has been generated for when_connected()
            new_d = self.factory.when_connected()
            assert new_d.called is False
            assert new_d != connected_d
            assert f.value.reason == error_msg

        self.factory.protocol = _get_protocol
        connected_d = self.factory.when_connected()
        connected_d.addCallbacks(lambda r: ValueError(f"This should fail but I got: {r!r}"), _check)
        self.factory.buildProtocol(None)
        return connected_d

    def test_when_connected_connectiondone(self):
        return self._test_when_connected_disconnected(
            ConnectionDone,
            "The TCP connection appears to have started, but the TLS or AMQP handshake "
            "with the broker failed; check your connection and authentication "
            "parameters and ensure your user has permission to access the vhost",
        )

    def test_when_connected_connectionlost(self):
        return self._test_when_connected_disconnected(
            ConnectionLost,
            "The network connection to the broker was lost in a non-clean fashion (%r);"
            " the connection should be restarted by Twisted.",
        )

    def test_when_connected_unexpected_failure(self):
        """Assert when_connected errbacks when the connection fails."""

        class DummyError(Exception):
            pass

        def _get_protocol(*a, **kw):
            protocol = mock.Mock(name="protocol mock")
            # Fail immediately
            protocol.ready = defer.fail(DummyError())
            return protocol

        def _check(f):
            f.trap(DummyError)
            # Make sure a new Deferred has been generated for when_connected()
            new_d = self.factory.when_connected()
            assert new_d.called is False
            assert new_d != connected_d

        self.factory.protocol = _get_protocol
        connected_d = self.factory.when_connected()
        connected_d.addCallbacks(lambda r: ValueError(f"This should fail but I got: {r!r}"), _check)
        with mock.patch("fedora_messaging.twisted.factory._std_log") as mock_log:
            self.factory.buildProtocol(None)
        mock_log.error.assert_called()
        last_log_call_args = mock_log.error.call_args_list[-1][0]
        assert last_log_call_args[0] == (
            "The connection failed with an unexpected exception; please report this bug: %s"
        )
        assert last_log_call_args[1].startswith("Traceback (most recent call last):")
        return connected_d

    def test_consume_anonymous(self):
        """Assert consume handles anonymous queues."""
        # Use server-generated queue names
        queue_config = {
            "queue": "",
            "durable": False,
            "auto_delete": True,
            "exclusive": True,
        }
        declared_queue = mock.Mock()
        self.protocol.declare_queue.side_effect = lambda q: declared_queue
        # Mock the consume call
        callback = mock.Mock()
        self.protocol.consume.side_effect = lambda cb, queue: defer.succeed(
            Consumer(queue=queue, callback=cb)
        )
        bindings = [{"exchange": "amq.topic", "routing_keys": ["#"]}]
        expected_bindings = [{"queue": declared_queue, "exchange": "amq.topic", "routing_key": "#"}]

        self.factory.buildProtocol(None)
        self.protocol.ready.callback(None)
        d = self.factory.when_connected()

        def _consume(_):
            return self.factory.consume(callback, bindings, {"": queue_config})

        def _check(_):
            assert len(self.factory._consumers) == 1
            consumer = self.factory._consumers[0].consumer
            assert consumer.queue == declared_queue
            assert consumer.callback == callback
            assert queue_config == self.factory._consumers[0].queue
            assert expected_bindings == self.factory._consumers[0].bindings

            self.protocol.declare_queue.assert_called_once_with(queue_config)
            self.protocol.bind_queues.assert_called_once_with(expected_bindings)
            self.protocol.consume.assert_called_once_with(callback, declared_queue)

        d.addCallback(_consume)
        d.addCallback(_check)
        return d

    def test_consume_anonymous_reconnect(self):
        """Assert consume handles reconnecting anonymous queues."""
        # Use server-generated queue names
        queue_config = {
            "queue": "",
            "durable": False,
            "auto_delete": True,
            "exclusive": True,
        }
        queue_orig = mock.Mock(name="queue_orig")
        queue_new = mock.Mock(name="queue_new")
        self.protocol.declare_queue.side_effect = lambda q: queue_new
        # Prepare the mocked existing consumer
        callback = mock.Mock()
        bindings = [{"exchange": "amq.topic", "routing_key": "#"}]
        expected_bindings = [{"queue": queue_new, "exchange": "amq.topic", "routing_key": "#"}]
        consumer = Consumer(queue=queue_orig, callback=callback)
        self.factory._consumers = [
            ConsumerRecord(consumer=consumer, queue=queue_config, bindings=bindings)
        ]

        self.factory.buildProtocol(None)
        self.protocol.ready.callback(None)
        d = self.factory.when_connected()

        def _check(_):
            assert len(self.factory._consumers) == 1
            self.protocol.declare_queue.assert_called_once_with(queue_config)
            self.protocol.bind_queues.assert_called_once_with(expected_bindings)
            self.protocol.consume.assert_called_once_with(callback, queue_new, consumer)

        d.addCallback(_check)
        return d


@pytest.mark.parametrize(
    "parameters,confirms,msg",
    [
        (
            pika.ConnectionParameters(),
            True,
            (
                "FedoraMessagingFactoryV2(parameters=<ConnectionParameters host=localhost"
                " port=5672 virtual_host=/ ssl=False>, confirms=True)"
            ),
        ),
        (
            pika.ConnectionParameters(
                host="example.com",
                credentials=pika.PlainCredentials("user", "secret"),
                port=5671,
                virtual_host="/pub",
            ),
            True,
            (
                "FedoraMessagingFactoryV2(parameters=<ConnectionParameters host=example.com"
                " port=5671 virtual_host=/pub ssl=False>, confirms=True)"
            ),
        ),
    ],
)
def test_repr(parameters, confirms, msg):
    """Assert __repr__ prints useful information"""
    f = FedoraMessagingFactoryV2(parameters, confirms)
    assert repr(f) == msg



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Tests for the :module:`fedora_messaging.cli` module."""


import errno
import os
import sys
from unittest import mock

import click
import pytest
import requests
from click.testing import CliRunner
from twisted.internet import error
from twisted.python import failure

from fedora_messaging import cli, config, exceptions, message, testing
from fedora_messaging.twisted import consumer


@pytest.fixture
def good_conf(fixtures_dir):
    return os.path.join(fixtures_dir, "good_conf.toml")


@pytest.fixture
def good_msg_dump(fixtures_dir):
    return os.path.join(fixtures_dir, "good_msg_dump.txt")


def echo(message):
    """Echo the message received to standard output."""
    print(str(message))


plain_object = object()


@mock.patch("fedora_messaging.config.conf.setup_logging", mock.Mock())
class TestBaseCli:
    """Unit tests for the base command of the CLI."""

    def test_no_conf(self):
        """Assert the CLI runs without exploding."""
        runner = CliRunner()
        result = runner.invoke(cli.cli)
        assert result.exit_code == 0


@mock.patch("fedora_messaging.cli.reactor", mock.Mock())
@mock.patch("fedora_messaging.config.conf.setup_logging", mock.Mock())
class TestConsumeCli:
    """Unit tests for the 'consume' command of the CLI."""

    def setup_method(self, method):
        self.runner = CliRunner()

    def teardown_method(self, method):
        """Make sure each test has a fresh default configuration."""
        config.conf = config.LazyConfig()
        config.conf.load_config(config_path="")

    @mock.patch("fedora_messaging.cli.api.twisted_consume")
    def test_good_conf(self, mock_consume, good_conf):
        """Assert providing a configuration file via the CLI works."""
        result = self.runner.invoke(cli.cli, ["--conf=" + good_conf, "consume"])
        mock_consume.assert_called_with(
            echo,
            bindings=[{"exchange": "e", "queue": "q", "routing_keys": ["#"]}],
            queues=config.conf["queues"],
        )
        assert 0 == result.exit_code

    @mock.patch("fedora_messaging.cli.api.twisted_consume")
    def test_conf_env_support(self, mock_consume, good_conf):
        """Assert FEDORA_MESSAGING_CONF environment variable is supported."""
        result = self.runner.invoke(cli.cli, ["consume"], env={"FEDORA_MESSAGING_CONF": good_conf})
        mock_consume.assert_called_with(
            echo,
            bindings=[{"exchange": "e", "queue": "q", "routing_keys": ["#"]}],
            queues=config.conf["queues"],
        )
        assert 0 == result.exit_code

    @mock.patch("fedora_messaging.cli.api.twisted_consume")
    def test_bad_conf(self, mock_consume, fixtures_dir):
        """Assert a bad configuration file is reported."""
        BAD_CONF = os.path.join(fixtures_dir, "bad_conf.toml")
        expected_err = (
            "Error: Invalid value: Configuration error: Failed to parse"
            f" {BAD_CONF}: Invalid value (at line 1, column 20)"
        )
        result = self.runner.invoke(cli.cli, ["--conf=" + BAD_CONF, "consume"])
        assert 2 == result.exit_code
        assert expected_err in result.output

    @mock.patch("fedora_messaging.cli.api.twisted_consume")
    def test_missing_conf(self, mock_consume):
        """Assert a missing configuration file is reported."""
        result = self.runner.invoke(cli.cli, ["--conf=thispathdoesnotexist", "consume"])
        assert 2 == result.exit_code
        assert "Error: Invalid value: thispathdoesnotexist is not a file" in result.output

    @mock.patch("fedora_messaging.cli.api.twisted_consume")
    def test_good_cli_bindings(self, mock_consume):
        """Assert providing a bindings via the CLI works."""
        config.conf["callback"] = "tests.unit.test_cli:echo"

        result = self.runner.invoke(
            cli.cli,
            [
                "consume",
                "--exchange=e",
                "--queue-name=qn",
                "--routing-key=rk1",
                "--routing-key=rk2",
            ],
        )
        mock_consume.assert_called_once_with(
            echo,
            bindings=[{"exchange": "e", "queue": "qn", "routing_keys": ("rk1", "rk2")}],
            queues={
                "qn": {
                    "durable": False,
                    "auto_delete": True,
                    "exclusive": True,
                    "arguments": {},
                }
            },
        )
        assert 0 == result.exit_code

    @mock.patch("fedora_messaging.cli.api.twisted_consume")
    def test_no_cli_bindings(self, mock_consume, good_conf):
        """Assert providing a bindings via configuration works."""
        expected_bindings = [{"exchange": "e", "queue": "q", "routing_keys": ["#"]}]

        result = self.runner.invoke(cli.cli, ["--conf=" + good_conf, "consume"])

        assert 0 == result.exit_code
        mock_consume.assert_called_once_with(
            echo, bindings=expected_bindings, queues=config.conf["queues"]
        )

    @mock.patch("fedora_messaging.cli.api.twisted_consume")
    def test_queue_and_routing_key(self, mock_consume):
        """Asser  providing improper bindings is reported."""
        config.conf["callback"] = "tests.unit.test_cli:echo"

        result = self.runner.invoke(cli.cli, ["consume", "--queue-name=qn", "--routing-key=rk"])

        mock_consume.assert_called_once_with(
            echo,
            bindings=[{"exchange": "amq.topic", "queue": "qn", "routing_keys": ("rk",)}],
            queues={
                "qn": {
                    "durable": False,
                    "auto_delete": True,
                    "exclusive": True,
                    "arguments": {},
                }
            },
        )
        assert 0 == result.exit_code

    @mock.patch("fedora_messaging.cli.api.twisted_consume")
    def test_good_cli_callable(self, mock_consume):
        """Assert providing a callable via the CLI works."""
        result = self.runner.invoke(cli.cli, ["consume", "--callback=tests.unit.test_cli:echo"])

        mock_consume.assert_called_once_with(
            echo, bindings=config.conf["bindings"], queues=config.conf["queues"]
        )
        assert 0 == result.exit_code

    @mock.patch("fedora_messaging.cli.importlib")
    @mock.patch("fedora_messaging.cli.api.twisted_consume")
    def test_app_name(self, mock_consume, mock_importlib, monkeypatch):
        """Assert provided app name is saved in config."""
        monkeypatch.setitem(config.conf, "bindings", "b")
        monkeypatch.setitem(config.conf, "queues", "c")
        cli_options = {"callback": "mod:callable", "app-name": "test_app_name"}
        mock_mod_with_callable = mock.Mock(spec=["callable"])
        mock_importlib.import_module.return_value = mock_mod_with_callable
        result = self.runner.invoke(
            cli.cli,
            [
                "consume",
                "--callback=" + cli_options["callback"],
                "--app-name=" + cli_options["app-name"],
            ],
        )
        assert config.conf["client_properties"]["app"] == cli_options["app-name"]
        mock_importlib.import_module.assert_called_once_with("mod")
        mock_consume.assert_called_once_with(
            mock_mod_with_callable.callable, bindings="b", queues="c"
        )
        assert 0 == result.exit_code

    @mock.patch("fedora_messaging.cli.importlib")
    @mock.patch("fedora_messaging.cli.api.twisted_consume")
    def test_missing_cli_and_conf_callable(self, mock_consume, mock_importlib, monkeypatch):
        """Assert missing callable via cli and in conf is reported."""
        monkeypatch.setitem(config.conf, "bindings", "b")
        monkeypatch.setitem(config.conf, "callback", None)
        mock_mod_with_callable = mock.Mock(spec=["callable"])
        mock_importlib.import_module.return_value = mock_mod_with_callable
        result = self.runner.invoke(cli.cli, ["consume"])
        mock_importlib.import_module.assert_not_called()
        mock_consume.assert_not_called()
        assert (
            "A Python path to a callable object that accepts the message must be provided"
            ' with the "--callback" command line option or in the configuration file'
            in result.output
        )
        assert 1 == result.exit_code

    @mock.patch("fedora_messaging.cli.importlib")
    @mock.patch("fedora_messaging.cli.api.twisted_consume")
    def test_cli_callable_wrong_format(self, mock_consume, mock_importlib, monkeypatch):
        """Assert a wrong callable format is reported."""
        cli_options = {"callback": "modcallable"}
        mock_mod_with_callable = mock.Mock(spec=["callable"])
        mock_importlib.import_module.return_value = mock_mod_with_callable
        result = self.runner.invoke(cli.cli, ["consume", "--callback=" + cli_options["callback"]])
        mock_importlib.import_module.assert_not_called()
        mock_consume.assert_not_called()
        assert (
            "Unable to parse the callback path ({}); the "
            'expected format is "my_package.module:'
            'callable_object"'.format(cli_options["callback"]) in result.output
        )
        assert 1 == result.exit_code

    @mock.patch("fedora_messaging.cli.importlib")
    @mock.patch("fedora_messaging.cli.api.twisted_consume")
    def test_cli_callable_import_failure_cli_opt(self, mock_consume, mock_importlib, monkeypatch):
        """Assert module with callable import failure is reported."""
        monkeypatch.setitem(config.conf, "bindings", "b")
        cli_options = {"callback": "mod:callable"}
        error_message = "No module named 'mod'"
        mock_importlib.import_module.side_effect = ImportError(error_message)
        result = self.runner.invoke(cli.cli, ["consume", "--callback=" + cli_options["callback"]])
        mock_importlib.import_module.assert_called_once_with("mod")
        mock_consume.assert_not_called()
        assert (
            f"Failed to import the callback module ({error_message}) provided in the --callback argument"
            in result.output
        )
        assert 1 == result.exit_code

    def test_cli_callable_import_failure_conf(self):
        """Assert module with callable import failure is reported."""
        config.conf["callback"] = "donotmakethismoduleorthetestbreaks:function"

        result = self.runner.invoke(cli.cli, ["consume"])

        assert (
            result.output == "Error: Failed to import the callback module "
            "(No module named 'donotmakethismoduleorthetestbreaks') "
            "provided in the configuration file\n"
        )
        assert 1 == result.exit_code

    @mock.patch("fedora_messaging.cli.getattr")
    @mock.patch("fedora_messaging.cli.importlib")
    @mock.patch("fedora_messaging.cli.api.twisted_consume")
    def test_callable_getattr_failure(
        self, mock_consume, mock_importlib, mock_getattr, monkeypatch
    ):
        """Assert finding callable in module failure is reported."""
        monkeypatch.setitem(config.conf, "bindings", "b")
        cli_options = {"callback": "mod:callable"}
        error_message = "module 'mod' has no attribute 'callable'"
        mock_mod_with_callable = mock.Mock(spec=["callable"])
        mock_importlib.import_module.return_value = mock_mod_with_callable
        mock_getattr.side_effect = AttributeError(error_message)
        result = self.runner.invoke(cli.cli, ["consume", "--callback=" + cli_options["callback"]])
        mock_importlib.import_module.assert_called_once_with("mod")
        mock_consume.assert_not_called()
        assert (
            "Unable to import {} ({}); is the package installed? The python path should "
            'be in the format "my_package.module:callable_object"'.format(
                cli_options["callback"], error_message
            )
            in result.output
        )
        assert 1 == result.exit_code

    def test_consume_improper_callback_object(self):
        """Assert improper callback object type failure is reported."""
        error_message = "Callback must be a class that implements __call__ or a function."

        result = self.runner.invoke(
            cli.cli,
            ["consume", "--callback=tests.unit.test_cli:plain_object"],
        )

        assert error_message in result.output
        assert 2 == result.exit_code


@mock.patch("fedora_messaging.cli.reactor")
class TestConsumeCallback:
    """Unit tests for the twisted_consume callback."""

    def test_callback(self, mock_reactor):
        """Assert when the last consumer calls back, the reactor stops."""
        consumers = (consumer.Consumer(), consumer.Consumer())
        cli._consume_callback(consumers)

        consumers[0].result.callback(consumers[0])
        assert 0 == mock_reactor.stop.call_count
        consumers[1].result.callback(consumers[1])
        assert 1 == mock_reactor.stop.call_count

    def test_callback_reactor_stopped(self, mock_reactor):
        """Assert already-stopped reactor is handled."""
        consumers = (consumer.Consumer(), consumer.Consumer())
        mock_reactor.stop.side_effect = error.ReactorNotRunning()

        cli._consume_callback(consumers)
        try:
            consumers[0].result.callback(consumers[0])
            consumers[1].result.callback(consumers[1])
            assert 1 == mock_reactor.stop.call_count
        except error.ReactorNotRunning:
            pytest.fail("ReactorNotRunning exception wasn't handled.")

    def test_errback_halt_consumer(self, mock_reactor):
        """Assert _exit_code is set with the HaltConsumer code."""
        consumers = (consumer.Consumer(),)
        e = exceptions.HaltConsumer()
        f = failure.Failure(e, exceptions.HaltConsumer)
        cli._consume_callback(consumers)

        consumers[0].result.errback(f)
        assert 1 == mock_reactor.stop.call_count
        assert 0 == cli._exit_code

    @mock.patch("fedora_messaging.cli._log")
    def test_errback_halt_consumer_nonzero(self, mock_log, mock_reactor):
        """Assert _exit_code is set with the HaltConsumer code and logged if non-zero"""
        consumers = (consumer.Consumer(),)
        e = exceptions.HaltConsumer(exit_code=42)
        f = failure.Failure(e, exceptions.HaltConsumer)
        cli._consume_callback(consumers)

        consumers[0].result.errback(f)
        assert 1 == mock_reactor.stop.call_count
        assert 42 == cli._exit_code
        mock_log.error.assert_called_with(
            "Consumer halted with non-zero exit code (%d): %s", 42, "None"
        )

    def test_errback_canceled(self, mock_reactor):
        """Assert exit code is 2 when the consumer is canceled."""
        consumers = (consumer.Consumer(),)
        e = exceptions.ConsumerCanceled()
        f = failure.Failure(e, exceptions.ConsumerCanceled)
        cli._consume_callback(consumers)

        consumers[0].result.errback(f)
        assert 1 == mock_reactor.stop.call_count
        assert 12 == cli._exit_code

    @mock.patch("fedora_messaging.cli._log")
    def test_errback_general_exception(self, mock_log, mock_reactor):
        """Assert exit code is 1 when an unexpected error occurs."""
        consumers = (consumer.Consumer(),)
        e = Exception("boom")
        f = failure.Failure(e, Exception)
        cli._consume_callback(consumers)

        consumers[0].result.errback(f)
        assert 1 == mock_reactor.stop.call_count
        assert 13 == cli._exit_code
        mock_log.error.assert_called_once_with(
            "Unexpected error occurred in consumer %r: %r", consumers[0], f
        )


class TestConsumeErrback:
    """Unit tests for the twisted_consume errback."""

    def setup_method(self, method):
        cli._exit_code = 0

    def teardown_method(self, method):
        cli._exit_code = 0

    @mock.patch("fedora_messaging.cli.reactor")
    def test_errback_permission(self, mock_reactor):
        """Assert permission exceptions are caught and exit with 15."""
        f = failure.Failure(exceptions.PermissionException("queue", "boop", "none"))

        cli._consume_errback(f)

        assert 15 == cli._exit_code

    @mock.patch("fedora_messaging.cli.reactor")
    def test_errback_bad_declaration(self, mock_reactor):
        """Assert declaration exceptions are caught and exit with 10."""
        f = failure.Failure(exceptions.BadDeclaration("queue", "boop", "none"))

        cli._consume_errback(f)

        assert 10 == cli._exit_code

    @mock.patch("fedora_messaging.cli.reactor")
    def test_errback_connection_exception(self, mock_reactor):
        """Assert connection exceptions are caught and exit with 14."""
        f = failure.Failure(exceptions.ConnectionException(reason="eh"))

        cli._consume_errback(f)

        assert 14 == cli._exit_code

    @mock.patch("fedora_messaging.cli.reactor")
    def test_errback_general_exception(self, mock_reactor):
        """Assert general exceptions are caught and exit with 11."""
        f = failure.Failure(Exception("boop"))

        cli._consume_errback(f)

        assert 11 == cli._exit_code


class CallbackFromFilesytem:
    """Unit tests for :func:`fedora_messaging.cli._callback_from_filesystem`."""

    def test_good_callback(self, fixtures_dir):
        """Assert loading a callback from a file works."""
        cb = cli._callback_from_filesystem(os.path.join(fixtures_dir, "callback.py") + ":rand")
        assert 4 == cb(None)

    def test_bad_format(self):
        """Assert an exception is raised if the format is bad."""
        with pytest.raises(click.ClickException) as cm:
            cli._callback_from_filesystem("file/with/no/function.py")

        assert (
            "Unable to parse the '--callback-file' option; the "
            'expected format is "path/to/file.py:callable_object" where '
            '"callable_object" is the name of the function or class in the '
            "Python file" == cm.value.message
        )

    def test_invalid_file(self, fixtures_dir):
        """Assert an exception is raised if the Python file can't be executed."""
        with pytest.raises(click.ClickException) as cm:
            cli._callback_from_filesystem(os.path.join(fixtures_dir, "bad_cb") + ":missing")

        if sys.version_info >= (3, 10) and sys.version_info < (3, 10, 4):
            # https://github.com/python/cpython/issues/90398
            exc_msg = "invalid syntax. Perhaps you forgot a comma?"
        else:
            exc_msg = "invalid syntax"
        assert (
            "The {} file raised the following exception during execution: "
            "{} (bad_cb, line 1)".format(
                os.path.join(fixtures_dir, "bad_cb"),
                exc_msg,
            )
            == cm.value.message
        )

    def test_callable_does_not_exist(self, fixtures_dir):
        """Assert an exception is raised if the callable is missing."""
        with pytest.raises(click.ClickException) as cm:
            cli._callback_from_filesystem(os.path.join(fixtures_dir, "callback.py") + ":missing")

        assert (
            "The 'missing' object was not found in the '{}' file."
            "".format(os.path.join(fixtures_dir, "callback.py")) == cm.value.message
        )

    def test_file_does_not_exist(self):
        """Assert an exception is raised if the file doesn't exist."""
        with pytest.raises(click.ClickException) as cm:
            cli._callback_from_filesystem("file/that/is/missing.py:callable")

        assert (
            "An IO error occurred: [Errno 2] No such file or directory: 'file/that/is/missing.py'"
            == cm.value.message
        )


@mock.patch("fedora_messaging.config.conf.setup_logging", mock.Mock())
class TestPublishCli:
    """Unit tests for the 'publish' command of the CLI."""

    def setup_method(self, method):
        self.runner = CliRunner()

    def teardown_method(self, method):
        """Make sure each test has a fresh default configuration."""
        config.conf = config.LazyConfig()
        config.conf.load_config()

    def test_correct_msg_in_file(self, good_conf, good_msg_dump):
        """Assert providing path to file with correct message via the CLI works."""
        cli_options = {"file": good_msg_dump, "exchange": "test_pe"}
        expected_msg = message.Message(
            body={"test_key1": "test_value1"}, topic="test_topic", severity=message.INFO
        )

        with testing.mock_sends(expected_msg):
            result = self.runner.invoke(
                cli.cli,
                [
                    "--conf=" + good_conf,
                    "publish",
                    "--exchange=" + cli_options["exchange"],
                    cli_options["file"],
                ],
            )
        assert "Publishing message with topic test_topic" in result.output
        assert 0 == result.exit_code

    @mock.patch("fedora_messaging.cli.api.publish")
    def test_file_with_corrupted_json(self, mock_publish, fixtures_dir, good_conf):
        """Assert providing path to file with corrupted message json via the CLI works."""
        wrong_json_msg_dump = os.path.join(fixtures_dir, "wrong_json_msg_dump.txt")
        cli_options = {"file": wrong_json_msg_dump, "exchange": "test_pe"}
        result = self.runner.invoke(
            cli.cli,
            [
                "--conf=" + good_conf,
                "publish",
                "--exchange=" + cli_options["exchange"],
                cli_options["file"],
            ],
        )
        assert "Error: Unable to validate message:" in result.output
        mock_publish.assert_not_called()
        assert 2 == result.exit_code

    @mock.patch("fedora_messaging.cli.api.publish")
    def test_file_with_msg_without_topic(self, mock_publish, fixtures_dir, good_conf):
        """Assert providing path to file with incorrect message via the CLI works."""
        msg_without_topic_dump = os.path.join(fixtures_dir, "msg_without_topic_dump.txt")
        cli_options = {"file": msg_without_topic_dump, "exchange": "test_pe"}
        result = self.runner.invoke(
            cli.cli,
            [
                "--conf=" + good_conf,
                "publish",
                "--exchange=" + cli_options["exchange"],
                cli_options["file"],
            ],
        )
        assert "Error: Unable to validate message: 'topic' is a required property" in result.output
        mock_publish.assert_not_called()
        assert 2 == result.exit_code

    @mock.patch("fedora_messaging.cli.api.publish")
    def test_file_with_invalid_msg(self, mock_publish, fixtures_dir, good_conf):
        """Assert providing path to file with incorrect message via the CLI works."""
        invalid_msg_dump = os.path.join(fixtures_dir, "invalid_msg_dump.txt")
        cli_options = {"file": invalid_msg_dump, "exchange": "test_pe"}
        result = self.runner.invoke(
            cli.cli,
            [
                "--conf=" + good_conf,
                "publish",
                "--exchange=" + cli_options["exchange"],
                cli_options["file"],
            ],
        )
        assert "Error: Unable to validate message: [] is not of type 'object'" in result.output
        mock_publish.assert_not_called()
        assert 2 == result.exit_code

    @mock.patch("fedora_messaging.cli.api.publish")
    def test_publish_rejected_message(self, mock_publish, good_conf, good_msg_dump):
        """Assert a rejected message is reported."""
        cli_options = {"file": good_msg_dump, "exchange": "test_pe"}
        error_message = "Message rejected"
        mock_publish.side_effect = exceptions.PublishReturned(error_message)
        result = self.runner.invoke(
            cli.cli,
            [
                "--conf=" + good_conf,
                "publish",
                "--exchange=" + cli_options["exchange"],
                cli_options["file"],
            ],
        )
        assert ("Unable to publish message: " + error_message) in result.output
        mock_publish.assert_called_once()
        assert errno.EREMOTEIO == result.exit_code

    @mock.patch("fedora_messaging.cli.api.publish")
    def test_publish_connection_failed(self, mock_publish, good_conf, good_msg_dump):
        """Assert a connection problem is reported."""
        cli_options = {"file": good_msg_dump, "exchange": "test_pe"}
        mock_publish.side_effect = exceptions.PublishTimeout(reason="timeout")
        result = self.runner.invoke(
            cli.cli,
            [
                "--conf=" + good_conf,
                "publish",
                "--exchange=" + cli_options["exchange"],
                cli_options["file"],
            ],
        )
        assert "Unable to connect to the message broker: timeout" in result.output
        mock_publish.assert_called_once()
        assert errno.ECONNREFUSED == result.exit_code

    @mock.patch("fedora_messaging.cli.api.publish")
    def test_publish_general_publish_error(self, mock_publish, good_conf, good_msg_dump):
        """Assert a connection problem is reported."""
        cli_options = {"file": good_msg_dump, "exchange": "test_pe"}
        mock_publish.side_effect = exceptions.PublishException(reason="eh")
        result = self.runner.invoke(
            cli.cli,
            [
                "--conf=" + good_conf,
                "publish",
                "--exchange=" + cli_options["exchange"],
                cli_options["file"],
            ],
        )
        assert "A general publish exception occurred: eh" in result.output
        mock_publish.assert_called_once()
        assert 1 == result.exit_code


@mock.patch("fedora_messaging.config.conf.setup_logging", mock.Mock())
class TestRecordCli:
    """Unit tests for the 'record' command of the CLI."""

    def setup_method(self, method):
        self.runner = CliRunner()

    def teardown_method(self, method):
        """Make sure each test has a fresh default configuration."""
        config.conf = config.LazyConfig()
        config.conf.load_config(config_path="")

    @mock.patch("fedora_messaging.cli._consume")
    def test_good_cli_bindings(self, mock_consume):
        """Assert arguments are forwarded to the _consume function."""
        cli_options = {
            "file": "test_file.txt",
            "exchange": "e",
            "queue-name": "qn",
            "routing-keys": ("rk1", "rk2"),
        }
        self.runner.invoke(
            cli.cli,
            [
                "record",
                cli_options["file"],
                "--exchange=" + cli_options["exchange"],
                "--queue-name=" + cli_options["queue-name"],
                "--routing-key=" + cli_options["routing-keys"][0],
                "--routing-key=" + cli_options["routing-keys"][1],
            ],
        )
        mock_consume.assert_called_once_with("e", "qn", ("rk1", "rk2"), mock.ANY, "recorder")


class TestRecorderClass:
    """Unit tests for the 'Recorder' class."""

    def test_save_recorded_messages_when_limit_is_reached(self):
        """Assert that collected messages are saved to file when limit is reached."""
        msg1 = message.Message(
            body={"test_key1": "test_value1"},
            topic="test_topic1",
            severity=message.INFO,
        )
        msg1._properties.headers["sent-at"] = "2018-11-18T10:11:41+00:00"
        msg1.id = "273ed91d-b8b5-487a-9576-95b9fbdf3eec"

        msg2 = message.Message(
            body={"test_key2": "test_value2"},
            topic="test_topic2",
            severity=message.INFO,
        )
        msg2._properties.headers["sent-at"] = "2018-11-18T10:11:41+00:00"
        msg2.id = "273ed91d-b8b5-487a-9576-95b9fbdf3eec"

        mock_file = mock.MagicMock()
        test_recorder = cli.Recorder(2, mock_file)
        test_recorder.collect_message(msg1)
        mock_file.write.assert_called_with(
            '{"body": {"test_key1": "test_value1"}, "headers"'
            ': {"fedora_messaging_schema": "base.message", "fedora_messaging_severity": 20, '
            '"priority": 0, "sent-at": "2018-11-18T10:11:41+00:00"}, '
            '"id": "273ed91d-b8b5-487a-9576-95b9fbdf3eec", '
            '"priority": 0, "queue": null, "topic": "test_topic1"}\n'
        )

        with pytest.raises(exceptions.HaltConsumer) as cm:
            test_recorder.collect_message(msg2)
        the_exception = cm.value
        assert the_exception.exit_code == 0
        assert test_recorder.counter == 2
        mock_file.write.assert_called_with(
            '{"body": {"test_key2": "test_value2"}, "headers": '
            '{"fedora_messaging_schema": "base.message", "fedora_messaging_severity": '
            '20, "priority": 0, "sent-at": "2018-11-18T10:11:41+00:00"}, "id": '
            '"273ed91d-b8b5-487a-9576-95b9fbdf3eec", "priority": 0, "queue": null, '
            '"topic": "test_topic2"}\n'
        )

    def test_recorded_messages_dumps_failed(self):
        """Assert that attempt to save improper recorded message is reported."""
        mock_file = mock.MagicMock()
        test_recorder = cli.Recorder(1, mock_file)
        with pytest.raises(exceptions.HaltConsumer) as cm:
            test_recorder.collect_message("msg1")
        the_exception = cm.value
        assert the_exception.exit_code == 1
        assert test_recorder.counter == 0
        mock_file.write.assert_not_called()


@mock.patch("fedora_messaging.config.conf.setup_logging", mock.Mock())
class TestReplayCli:
    """Unit tests for the 'replay' command of the CLI."""

    def setup_method(self, method):
        """Setup test method environment."""
        self.runner = CliRunner()

    def teardown_method(self, method):
        """Reset configuration after each test."""
        config.conf = config.LazyConfig()
        config.conf.load_config()

    @mock.patch("fedora_messaging.cli._get_message")
    @mock.patch("fedora_messaging.api.publish")
    def test_successful_message_replay(self, mock_publish, mock_get_message, monkeypatch):
        """Test successful replay of a message."""
        monkeypatch.setitem(config.conf, "topic_prefix", "dummy.topic.prefix")
        message_id = "123"
        datagrepper_url = "http://example.com"

        message_data = {
            "topic": "test.topic",
            "body": {"data_key": "data_value"},
        }
        mock_get_message.return_value = message_data

        result = self.runner.invoke(
            cli.replay, [message_id, f"--datagrepper-url={datagrepper_url}"]
        )

        assert result.exit_code == 0, f"Command did not exit as expected. Output: {result.output}"
        assert "has been successfully replayed" in result.output
        mock_publish.assert_called_once()
        published_msg = mock_publish.call_args[0][0]
        assert published_msg.topic == message_data["topic"]
        assert published_msg.body == message_data["body"]
        assert config.conf["topic_prefix"] == ""

    @mock.patch("fedora_messaging.cli.requests.get", side_effect=requests.HTTPError)
    def test_datagrepper_http_error(self, mock_get):
        """Test handling of HTTP errors when fetching message data."""
        message_id = "123"
        result = self.runner.invoke(cli.replay, [message_id])
        assert "Failed to retrieve message from Datagrepper" in result.output
        assert result.exit_code != 0

    @mock.patch("fedora_messaging.cli._get_message", return_value={"some": "data"})
    @mock.patch("fedora_messaging.cli.api.publish", side_effect=Exception("Publish failure"))
    def test_publish_failure(self, mock_publish, mock_get_message):
        """Test handling of exceptions during message publishing."""
        message_id = "123"
        result = self.runner.invoke(cli.replay, [message_id])
        assert result.exit_code != 0



#!/usr/bin/env python3

import importlib
import os
import site
import sys
import tempfile
import venv
from collections import defaultdict
from dataclasses import dataclass
from importlib.metadata import entry_points
from subprocess import run
from textwrap import dedent
from urllib.parse import urljoin

from sphinx.ext.napoleon.docstring import GoogleDocstring


SCHEMAS_FILE = "schema-packages.txt"
DOC_FILE = "user-guide/schemas.rst"
HEADER = """
=================
Available Schemas
=================

.. This file is autogenerated by the build-schemas-list.py script. Do not edit manually.

These are the topics that you can expect to see on Fedora's message bus,
sorted by the python package that contains their schema.
Install the corresponding python package if you want to make use of the schema
and access additional information on the message you're receiving.

In the Fedora Infrastructure, some of those topics will be prefixed by
``org.fedoraproject.stg.`` in staging and ``org.fedoraproject.prod.`` in production.
"""
DATAGREPPER_URL = "https://apps.fedoraproject.org/datagrepper/"
PREFIXES = ("org.fedoraproject.", "org.release-monitoring.")


def read_packages(schemas_filepath):
    packages = []
    with open(schemas_filepath) as schemas_file:
        for line in schemas_file:
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            packages.append(line)
    packages.sort()
    return packages


@dataclass
class Schema:
    topic: str
    package: str
    app_name: str
    category: str
    doc: str


def create_venv(dirname):
    print("Creating virtualenv...")
    venv.create(dirname, with_pip=True)
    activate_venv(dirname)


def activate_venv(dirname):
    # Remove system site-packages from the path
    for spdir in site.getsitepackages():
        if spdir in sys.path:
            sys.path.remove(spdir)
    sys.prefix = sys.exec_prefix = dirname
    site.addsitepackages(set(sys.path), [dirname])
    site.PREFIXES = [dirname]
    site.ENABLE_USER_SITE = False
    os.environ["VIRTUAL_ENV"] = dirname
    importlib.invalidate_caches()


def install_packages(dirname, packages):
    # Don't use pip as a library:
    # https://pip.pypa.io/en/stable/user_guide/#using-pip-from-your-program
    pip = os.path.join(dirname, "bin", "pip")
    print("Upgrading pip...")
    run([pip, "-q", "install", "--upgrade", "pip"], check=True)  # noqa: S603
    for package in packages:
        print(f"Installing {package}...")
        run([pip, "-q", "install", package], check=True)  # noqa: S603


def extract_docstring(cls):
    if not cls.__doc__:
        return None
    gds = GoogleDocstring(dedent(cls.__doc__), obj=cls)
    doc = []
    for line in gds.lines():
        if line.startswith(".. "):
            break
        if not line:
            continue
        doc.append(line)
    return " ".join(doc)


def get_schemas():
    schemas = defaultdict(list)
    for entry_point in entry_points(group="fedora.messages"):
        msg_cls = entry_point.load()
        if not msg_cls.topic:
            target = f"{entry_point.module}:{entry_point.attr}"
            if target != "fedora_messaging.message:Message":
                print(f"The {target} schema has no declared topic, skipping.")
            continue
        if msg_cls.deprecated:
            continue
        package_name = entry_point.dist.name
        doc = extract_docstring(msg_cls)
        category = _get_category(msg_cls.topic)
        try:
            app_name = msg_cls().app_name
        except Exception:
            # Sometimes we can't instantiate schema classes without an actual body
            app_name = None
        schemas[package_name].append(
            Schema(
                topic=msg_cls.topic,
                package=package_name,
                doc=doc,
                category=category,
                app_name=app_name,
            )
        )
    return schemas


def _is_prefixed(topic):
    return any(topic.startswith(prefix) for prefix in PREFIXES)


def _get_app_name(schemas):
    for schema in schemas:
        if schema.app_name is not None:
            return schema.app_name
    return None


def _get_category(topic):
    if _is_prefixed(topic):
        index = 3
    else:
        index = 0
    return topic.split(".")[index]


def write_doc(schemas, doc_filepath):
    with open(doc_filepath, "w") as doc_file:
        doc_file.write(HEADER)
        for package_name in sorted(schemas):
            package_schemas = schemas[package_name]
            app_name = _get_app_name(package_schemas)
            category = package_schemas[0].category
            title = app_name or category
            print(f"\n\n{title}", file=doc_file)
            print("=" * len(title), end="\n\n", file=doc_file)
            history_url = urljoin(DATAGREPPER_URL, f"raw?category={category}")
            print(
                f"You can view the history of `all {title} messages <{history_url}>`__ "
                "in datagrepper.\n",
                file=doc_file,
            )
            for schema in package_schemas:
                prod_topic = (
                    schema.topic
                    if _is_prefixed(schema.topic)
                    else f"org.fedoraproject.prod.{schema.topic}"
                )
                history_url = urljoin(DATAGREPPER_URL, f"raw?topic={prod_topic}")
                if schema.doc:
                    print(
                        f"* ``{schema.topic}``: {schema.doc.strip()} (`history <{history_url}>`__)",
                        file=doc_file,
                    )
                else:
                    print(
                        f"* ``{schema.topic}`` (`history <{history_url}>`__)",
                        file=doc_file,
                    )


def main():
    here = os.path.dirname(__file__)
    schemas_file = os.path.normpath(os.path.join(here, SCHEMAS_FILE))
    packages = read_packages(schemas_file)
    with tempfile.TemporaryDirectory() as tmpdirname:
        create_venv(tmpdirname)
        install_packages(tmpdirname, packages)
        schemas = get_schemas()
    doc_file = os.path.normpath(os.path.join(here, DOC_FILE))
    write_doc(schemas, doc_file)
    print(f"Wrote the documentation in {doc_file}")


if __name__ == "__main__":
    main()



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Once you've written code to publish or consume messages, you'll probably want
to test it. The :mod:`fedora_messaging.testing` module has utilities for common
test patterns.

If you find yourself implementing a pattern over and over in your test code,
consider contributing it here!
"""

import inspect
from contextlib import contextmanager
from unittest import mock


@contextmanager
def mock_sends(*expected_messages):
    """
    Assert a block of code results in the provided messages being sent without
    actually sending them.

    This is intended for unit tests. The call to publish is mocked out and messages
    are captured and checked at the end of the ``with``.

    For example:

        >>> from fedora_messaging import api, testing
        >>> def publishes():
        ...     api.publish(api.Message(body={"Hello": "world"}))
        ...
        >>> with testing.mock_sends(api.Message, api.Message(body={"Hello": "world"})):
        ...     publishes()
        ...     publishes()
        ...
        >>> with testing.mock_sends(api.Message(body={"Goodbye": "everybody"})):
        ...     publishes()
        ...
        AssertionError

    Args:
        *expected_messages: The messages you expect to be sent. These can be classes
            instances of classes derived from :class:`fedora_messaging.message.Message`.
            If the class is provided, the message is checked to make sure it is an
            instance of that class and that it passes schema validation. If an instance
            is provided, it is checked for equality with the sent message.

    Raises:
        AssertionError: If the messages published don't match the messages asserted.
    """

    sent = []
    with mock.patch("fedora_messaging.api.crochet"):
        with mock.patch("fedora_messaging.api._twisted_publish_wrapper") as mock_pub:
            yield sent

    messages = [call[0][0] for call in mock_pub.call_args_list]
    sent.extend(messages)
    if len(expected_messages) != len(messages):
        raise AssertionError(
            f"Expected {len(expected_messages)} messages to be sent, but {len(messages)} were sent"
        )
    for msg, expected in zip(messages, expected_messages):
        if inspect.isclass(expected):
            if not isinstance(msg, expected):
                raise AssertionError(
                    f"Expected message of type {expected}, but {msg.__class__} was sent"
                )
        else:
            assert msg.topic == expected.topic
            assert msg.body == expected.body
            assert msg == expected
        msg.validate()



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from urllib import parse

import pytest

from fedora_messaging import schema_utils


class TestAvatarURL:
    def test_base(self):
        expected = (
            "https://seccdn.libravatar.org/avatar/"
            "7cebe82e9588dcccabc7f32d50901befeb4dfa42baf08044791feea4b48b2e34"
            "?s=64&d=retro"
        )
        assert schema_utils.user_avatar_url("testuser") == expected

    def test_size(self):
        # Check default size == 64
        default_url = parse.urlparse(schema_utils.user_avatar_url("testuser"))
        assert ("s", "64") in parse.parse_qsl(default_url.query)
        # Check a different size
        custom_url = parse.urlparse(schema_utils.user_avatar_url("testuser", size=128))
        assert ("s", "128") in parse.parse_qsl(custom_url.query)

    def test_different_default(self):
        url = schema_utils.user_avatar_url("testuser", default="testdefault")
        assert ("d", "testdefault") in parse.parse_qsl(parse.urlparse(url).query)


class TestLibravatarURL:
    def test_openid(self):
        expected = (
            "https://seccdn.libravatar.org/avatar/"
            "ae5deb822e0d71992900471a7199d0d95b8e7c9d05c40a8245a281fd2c1d6684"
            "?s=64&d=retro"
        )
        assert schema_utils.libravatar_url(openid="testuser") == expected

    def test_email(self):
        expected = (
            "https://seccdn.libravatar.org/avatar/"
            "ae5deb822e0d71992900471a7199d0d95b8e7c9d05c40a8245a281fd2c1d6684"
            "?s=64&d=retro"
        )
        assert schema_utils.libravatar_url(email="testuser") == expected

    def test_invalid_params(self):
        with pytest.raises(ValueError):
            schema_utils.libravatar_url()



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Tests for :mod:`fedora_messaging.example`."""

from io import StringIO
from unittest import mock

from fedora_messaging import api, example


def test_printer():
    """Assert the printer callback prints messages."""
    message = api.Message(body="Hello world", topic="hi")
    message._headers = {
        "fedora_messaging_schema": "fedora_messaging.message:Message",
        "sent-at": "2019-07-30T19:12:22+00:00",
    }
    message.id = "95383db8-8cdc-4464-8276-d482ac28b0b6"
    expected_stdout = (
        "Id: 95383db8-8cdc-4464-8276-d482ac28b0b6\n"
        "Topic: hi\n"
        "Headers: {\n"
        '    "fedora_messaging_schema": "fedora_messaging.message:Message",\n'
        '    "sent-at": "2019-07-30T19:12:22+00:00"\n'
        "}\n"
        'Body: "Hello world"\n'
    )

    with mock.patch("sys.stdout", new_callable=StringIO) as mock_stdout:
        example.printer(message)

    assert expected_stdout == mock_stdout.getvalue()



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
The core Twisted interface, a protocol represent a specific connection to the
AMQP broker.

The :class:`.FedoraMessagingProtocolV2` inherits the
:class:`pika.adapters.twisted_connection.TwistedProtocolConnection` class and
adds a few additional methods.

When combined with the :class:`fedora_messaging.twisted.factory.FedoraMessagingFactoryV2`
class, it's easy to create AMQP consumers that last across connections.

For an overview of Twisted clients, see the `Twisted client documentation
<https://twistedmatrix.com/documents/current/core/howto/clients.html#protocol>`_.
"""


import logging

import pika
from pika.adapters.twisted_connection import TwistedProtocolConnection
from twisted.internet import defer, error

from .. import config
from ..exceptions import (
    BadDeclaration,
    ConnectionException,
    NoFreeChannels,
    PublishForbidden,
    PublishReturned,
)
from .consumer import Consumer


_std_log = logging.getLogger(__name__)


class FedoraMessagingProtocolV2(TwistedProtocolConnection):
    """A Twisted Protocol for the Fedora Messaging system.

    This protocol builds on the generic pika AMQP protocol to add calls
    specific to the Fedora Messaging implementation.

    Attributes:
        factory: The :class:`Factory` object that created this protocol. This
        is set by the factory that creates this protocol.

    Args:
        parameters (pika.ConnectionParameters): The connection parameters.
        confirms (bool): If True, all outgoing messages will require a
            confirmation from the server, and the Deferred returned from
            the publish call will wait for that confirmation.
    """

    name = "FedoraMessaging:Protocol"

    def __init__(self, parameters, confirms=True):
        TwistedProtocolConnection.__init__(self, parameters)
        self._confirms = confirms
        self._channel = None
        self._publish_channel = None
        # Map queue names to fedora_messaging.twisted.consumer.Consumer objects
        self._consumers = {}
        self.factory = None

    @defer.inlineCallbacks
    def _allocate_channel(self):
        """
        Allocate a new AMQP channel.

        Raises:
            NoFreeChannels: If this connection has reached its maximum number of channels.
            ConncetionException: If this connection is already closed.
        """
        try:
            channel = yield self.channel()
        except pika.exceptions.NoFreeChannels as e:
            raise NoFreeChannels() from e
        except pika.exceptions.ConnectionWrongStateError as e:
            raise ConnectionException(reason=e) from e
        _std_log.debug("Created AMQP channel id %d", channel.channel_number)
        if self._confirms:
            yield channel.confirm_delivery()
        defer.returnValue(channel)

    @defer.inlineCallbacks
    def connectionReady(self, res=None):
        """
        Callback invoked when the AMQP connection is ready (when self.ready fires).

        This API is not meant for users.

        Args:
            res: This is an unused argument that provides compatibility with Pika
                versions lower than 1.0.0.
        """
        self._channel = yield self._allocate_channel()

    @defer.inlineCallbacks
    def publish(self, message, exchange):
        """
        Publish a :class:`fedora_messaging.message.Message` to an `exchange`_
        on the message broker.

        Args:
            message (message.Message): The message to publish.
            exchange (str): The name of the AMQP exchange to publish to

        Raises:
            NoFreeChannels: If there are no available channels on this connection.
                If this occurs, you can either reduce the number of consumers on this
                connection or create an additional connection.
            PublishReturned: If the broker rejected the message. This can happen if
                there are resource limits that have been reached (full disk, for example)
                or if the message will be routed to 0 queues and the exchange is set to
                reject such messages.
            PublishForbidden: If the broker rejected the message because of permission
                issues (for example: the topic is not allowed).

        .. _exchange: https://www.rabbitmq.com/tutorials/amqp-concepts.html#exchanges
        """
        message.validate()
        if self._publish_channel is None:
            self._publish_channel = yield self._allocate_channel()

        try:
            yield self._publish_channel.basic_publish(
                exchange=exchange,
                routing_key=message._encoded_routing_key,
                body=message._encoded_body,
                properties=message._properties,
            )
        except (pika.exceptions.NackError, pika.exceptions.UnroutableError) as e:
            _std_log.error("Message was rejected by the broker (%s)", str(e))
            raise PublishReturned(reason=e) from e
        except pika.exceptions.ProbableAccessDeniedError as e:
            _std_log.error("Message was forbidden by the broker (%s)", str(e))
            raise PublishForbidden(reason=e) from e
        except (
            pika.exceptions.ChannelClosed,
            pika.exceptions.ChannelWrongStateError,
        ) as e:
            # Channel has been closed, we'll need to re-allocate it.
            self._publish_channel = None
            # Handle Forbidden errors
            if isinstance(e, pika.exceptions.ChannelClosed) and e.reply_code == 403:
                _std_log.error(
                    "Message was forbidden by the broker: %s",
                    e.reply_text,
                )
                raise PublishForbidden(reason=e) from e
            # In other cases, try to publish again
            yield self.publish(message, exchange)
        except (
            pika.exceptions.ConnectionClosed,
            error.ConnectionLost,
            error.ConnectionDone,
        ) as e:
            raise ConnectionException(reason=e) from e

    @defer.inlineCallbacks
    def consume(self, callback, queue, previous_consumer=None):
        """
        Register a message consumer that executes the provided callback when
        messages are received.

        The queue must exist prior to calling this method. If a consumer
        already exists for the given queue, the callback is simply updated and
        any new messages for that consumer use the new callback.

        Args:
            callback (callable): The callback to invoke when a message is received.
            queue (str): The name of the queue to consume from.
            previous_consumer (Consumer): If this is the resumption of a prior
                consumer, you can provide the previous consumer so its result
                deferred can be re-used.

        Returns:
            Deferred: A Deferred that fires when the consumer is successfully
                registered with the message broker. The callback receives a
                :class:`.Consumer` object that represents the AMQP consumer.
                The Deferred may error back with a :class:`PermissionException`
                if the user cannot read from the queue, a
                :class:`NoFreeChannels` if this connection has hit its channel
                limit, or a :class:`ConnectionException` if the connection dies
                before the consumer is successfully registered.

        NoFreeChannels: If there are no available channels on this connection.
            If this occurs, you can either reduce the number of consumers on this
            connection or create an additional connection.
        """
        if queue in self._consumers:
            self._consumers[queue].callback = callback
            defer.returnValue(self._consumers[queue])

        if previous_consumer is not None:
            consumer = previous_consumer
            # The queue name may have changed, especially in the case of server-generated queues.
            consumer.queue = queue
        else:
            consumer = Consumer(queue=queue, callback=callback)
        consumer._protocol = self
        consumer._channel = yield self._allocate_channel()

        yield consumer.consume()

        self._consumers[queue] = consumer
        _std_log.info("Successfully registered AMQP consumer %r", consumer)
        defer.returnValue(consumer)

    @defer.inlineCallbacks
    def declare_exchanges(self, exchanges):
        """
        Declare a number of exchanges at once.

        This simply wraps the :meth:`pika.channel.Channel.exchange_declare`
        method and deals with error handling and channel allocation.

        Args:
            exchanges (list of dict): A list of dictionaries, where each dictionary
                represents an exchange. Each dictionary can have the following keys:

                  * exchange (str): The exchange's name
                  * exchange_type (str): The type of the exchange ("direct", "topic", etc)
                  * passive (bool): If true, this will just assert that the exchange exists,
                    but won't create it if it doesn't. Defaults to the configuration value
                    :ref:`conf-passive-declares`
                  * durable (bool): Whether or not the exchange is durable
                  * arguments (dict): Extra arguments for the exchange's creation.
        Raises:
            NoFreeChannels: If there are no available channels on this connection.
                If this occurs, you can either reduce the number of consumers on this
                connection or create an additional connection.
            BadDeclaration: If an exchange could not be declared. This can occur
                if the exchange already exists, but does its type does not match
                (e.g. it is declared as a "topic" exchange, but exists as a "direct"
                exchange). It can also occur if it does not exist, but the current
                user does not have permissions to create the object.
        """
        channel = yield self._allocate_channel()
        try:
            for exchange in exchanges:
                args = exchange.copy()
                args.setdefault("passive", config.conf["passive_declares"])
                try:
                    yield channel.exchange_declare(**args)
                except pika.exceptions.ChannelClosed as e:
                    raise BadDeclaration("exchange", args, e) from e
        finally:
            try:
                channel.close()
            except pika.exceptions.AMQPError:
                pass  # pika doesn't handle repeated closes gracefully

    @defer.inlineCallbacks
    def declare_queues(self, queues):
        """
        Declare a list of queues.

        Args:
            queues (list of dict): A list of dictionaries, where each dictionary
                represents an exchange. Each dictionary can have the following keys:

                  * queue (str): The name of the queue
                  * passive (bool): If true, this will just assert that the queue exists,
                    but won't create it if it doesn't. Defaults to the configuration value
                    :ref:`conf-passive-declares`
                  * durable (bool): Whether or not the queue is durable
                  * exclusive (bool): Whether or not the queue is exclusive to this connection.
                  * auto_delete (bool): Whether or not the queue should be automatically
                    deleted once this connection ends.
                  * arguments (dict): Additional arguments for the creation of the queue.
        Raises:
            NoFreeChannels: If there are no available channels on this connection.
                If this occurs, you can either reduce the number of consumers on this
                connection or create an additional connection.
            BadDeclaration: If a queue could not be declared. This can occur
                if the queue already exists, but does its type does not match
                (e.g. it is declared as a durable queue, but exists as a non-durable
                queue). It can also occur if it does not exist, but the current
                user does not have permissions to create the object.
        """
        channel = yield self._allocate_channel()
        result_queues = []
        try:
            for queue in queues:
                args = queue.copy()
                args.setdefault("passive", config.conf["passive_declares"])
                try:
                    frame = yield channel.queue_declare(**args)
                except pika.exceptions.ChannelClosed as e:
                    raise BadDeclaration("queue", args, e) from e
                result_queues.append(frame.method.queue)
        finally:
            try:
                channel.close()
            except pika.exceptions.AMQPError:
                pass  # pika doesn't handle repeated closes gracefully
        defer.returnValue(result_queues)

    @defer.inlineCallbacks
    def declare_queue(self, queue):
        """
        Declare a queue. This is a convenience method to call :meth:`declare_queues` with a single
        argument.
        """
        names = yield self.declare_queues([queue])
        defer.returnValue(names[0])

    @defer.inlineCallbacks
    def bind_queues(self, bindings):
        """
        Declare a set of bindings between queues and exchanges.

        Args:
            bindings (list of dict): A list of binding definitions. Each dictionary
                must contain the "queue" key whose value is the name of the queue
                to create the binding on, as well as the "exchange" key whose value
                should be the name of the exchange to bind to. Additional acceptable
                keys are any keyword arguments accepted by
                :meth:`pika.channel.Channel.queue_bind`.

        Raises:
            NoFreeChannels: If there are no available channels on this connection.
                If this occurs, you can either reduce the number of consumers on this
                connection or create an additional connection.
            BadDeclaration: If a binding could not be declared. This can occur if the
                queue or exchange don't exist, or if they do, but the current user does
                not have permissions to create bindings.
        """
        channel = yield self._allocate_channel()
        try:
            for binding in bindings:
                try:
                    yield channel.queue_bind(**binding)
                except pika.exceptions.ChannelClosed as e:
                    raise BadDeclaration("binding", binding, e) from e
        finally:
            try:
                channel.close()
            except pika.exceptions.AMQPError:
                pass  # pika doesn't handle repeated closes gracefully

    @defer.inlineCallbacks
    def halt(self):
        """
        Signal to consumers they should stop after finishing any messages
        currently being processed, then close the connection.

        Returns:
            defer.Deferred: fired when all consumers have successfully stopped
            and the connection is closed.
        """
        if self.is_closed:
            # We were asked to stop because the connection is already gone.
            # There's no graceful way to stop because we can't acknowledge
            # messages in the middle of being processed.
            _std_log.info("Disconnect requested, but AMQP connection already gone")
            self._channel = None
            return

        _std_log.info(
            "Waiting for %d consumer(s) to finish processing before halting",
            len(self._consumers),
        )
        pending_cancels = []
        for c in list(self._consumers.values()):
            pending_cancels.append(c.cancel())
        yield defer.gatherResults(pending_cancels)
        _std_log.info("Finished canceling %d consumers", len(self._consumers))

        try:
            yield self.close()
        except pika.exceptions.ConnectionWrongStateError:
            pass  # Already closing, not a problem since that's what we want.
        self._consumers = {}
        self._channel = None

    def _forget_consumer(self, queue):
        """Forget about a consumer so it does not restart later.

        Args:
            queue (str): Forget consumers consuming from this queue.
        """
        # If client and server are racing to cancel it might already be gone which
        # is why both are marked as no cover.
        try:
            del self._consumers[queue]
        except KeyError:  # pragma: no cover
            pass
        self.factory._forget_consumer(queue)



# This file is part of fedora_messaging.
# Copyright (C) 2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import asyncio
import logging
import uuid

import pika
from twisted.internet import defer, error, reactor, threads
from twisted.python.failure import Failure

from .. import config
from ..exceptions import (
    ConnectionException,
    ConsumerCanceled,
    Drop,
    HaltConsumer,
    Nack,
    PermissionException,
    ValidationError,
)
from ..message import get_message


_std_log = logging.getLogger(__name__)


def _add_timeout(deferred, timeout):
    """
    Add a timeout to the given deferred. This is designed to work with both old
    Twisted and versions of Twisted with the addTimeout API. This is
    exclusively to support EL7.

    The deferred will errback with a :class:`defer.CancelledError` if the
    version of Twisted being used doesn't have the
    ``defer.Deferred.addTimeout`` API, otherwise it will errback with the
    normal ``error.TimeoutError``
    """
    try:
        deferred.addTimeout(timeout, reactor)
    except AttributeError:
        # Twisted 12.2 (in EL7) does not have the addTimeout API, so make do with
        # the slightly more annoying approach of scheduling a call to cancel which
        # is then canceled if the deferred succeeds before the timeout is up.
        delayed_cancel = reactor.callLater(timeout, deferred.cancel)

        def cancel_cancel_call(result):
            """Halt the delayed call to cancel if the deferred fires before the timeout."""
            if not delayed_cancel.called:
                delayed_cancel.cancel()
            return result

        deferred.addBoth(cancel_cancel_call)


def is_coro(func_or_obj):
    """Tests if a function is a coroutine function or a callable coroutine object."""
    # Until Python 3.10, inspect.iscoroutinefunction() will fail to identify AsyncMocks
    # as coroutines: https://github.com/python/cpython/issues/84753
    # Use asyncio.iscoroutinefunction() instead.
    return asyncio.iscoroutinefunction(func_or_obj) or (
        callable(func_or_obj) and asyncio.iscoroutinefunction(func_or_obj.__call__)
    )


class Consumer:
    """
    Represents a Twisted AMQP consumer and is returned from the call to
    :func:`fedora_messaging.api.twisted_consume`.

    Attributes:
        queue (str): The AMQP queue this consumer is subscribed to.
        callback (callable): The callback to run when a message arrives.
        result (twisted.internet.defer.Deferred):
            A deferred that runs the callbacks if the consumer exits gracefully
            after being canceled by a call to :meth:`Consumer.cancel` and
            errbacks if the consumer stops for any other reason. The reasons a
            consumer could stop are: a
            :class:`fedora_messaging.exceptions.PermissionExecption` if the
            consumer does not have permissions to read from the queue it is
            subscribed to, a :class:`.HaltConsumer` is raised by the consumer
            indicating it wishes to halt, an unexpected :class:`Exception` is
            raised by the consumer, or if the consumer is canceled by the
            server which happens if the queue is deleted by an administrator or
            if the node the queue lives on fails.
    """

    def __init__(self, queue=None, callback=None):
        self.queue = queue
        self.callback = callback
        self.result = defer.Deferred()

        # The current channel used by this consumer.
        self._channel = None
        # The unique ID for the AMQP consumer.
        self._tag = str(uuid.uuid4())
        # Used in the consumer read loop to know when it's being canceled.
        self._running = True
        # The current read loop
        self._read_loop = None
        # The protocol that currently runs this consumer, used when cancel is
        # called to remove itself from the protocol and its factory so it doesn't
        # restart on the next connection.
        self._protocol = None

    def __repr__(self):
        return f"Consumer(queue={self.queue}, callback={self.callback})"

    @defer.inlineCallbacks
    def consume(self):
        yield self._channel.basic_qos(
            prefetch_count=config.conf["qos"]["prefetch_count"],
            prefetch_size=config.conf["qos"]["prefetch_size"],
        )
        try:
            queue_object, _ = yield self._channel.basic_consume(
                queue=self.queue, consumer_tag=self._tag
            )
        except pika.exceptions.ChannelClosed as exc:
            if exc.args[0] == 403:
                raise PermissionException(
                    obj_type="queue", description=self.queue, reason=exc.args[1]
                ) from exc
            else:
                raise ConnectionException(reason=exc) from exc

        try:
            self._channel.add_on_cancel_callback(self._on_cancel_callback)
        except AttributeError:
            pass  # pika 1.0.0+

        self._read_loop = self._read(queue_object)
        self._read_loop.addErrback(self._read_loop_errback)

    @defer.inlineCallbacks
    def _read(self, queue_object):
        """
        The loop that reads from the message queue and calls the callback
        wrapper.

        Serialized Processing
        ---------------------
        This loop processes messages serially. This is because a second
        ``queue_object.get()`` operation can only occur after the Deferred from
        the callback completes. Thus, we can be sure that callbacks
        never run concurrently in two different threads.

        This is done rather than saturating the Twisted thread pool as the
        documentation for callbacks (in fedmsg and here) has never indicated
        that they are not thread-safe. In the future we can add a flag for users
        who are confident in their ability to write thread-safe code.

        Gracefully Halting
        ------------------
        This is a loop that only exits when the ``self._running`` variable is
        set to False. The call to cancel will set this to false, and will then
        wait for the Deferred from this function to call back in order to
        ensure the message finishes processing.

        The Deferred object only completes when this method returns, so we need
        to periodically check the status of ``self._running``. That's why
        there's a short timeout on the call to ``queue_object.get``.

        queue_object (pika.adapters.twisted_connection.ClosableDeferredQueue):
            The AMQP queue the consumer is bound to.
        """
        while self._running:
            yield self._read_one(queue_object)

    @defer.inlineCallbacks
    def _read_one(self, queue_object):
        try:
            deferred_get = queue_object.get()
            _add_timeout(deferred_get, 1)
            channel, delivery_frame, properties, body = yield deferred_get
        except (defer.TimeoutError, defer.CancelledError):
            return

        _std_log.debug(
            "Message arrived with delivery tag %s for %r",
            delivery_frame.delivery_tag,
            self._tag,
        )
        try:
            message = get_message(delivery_frame.routing_key, properties, body)
            message.queue = self.queue
        except ValidationError:
            _std_log.warning(
                "Message id %s did not pass validation; ignoring message",
                properties.message_id,
            )
            yield channel.basic_nack(delivery_tag=delivery_frame.delivery_tag, requeue=False)
            return

        try:
            _std_log.info(
                "Consuming message from topic %s (message id %s)",
                message.topic,
                properties.message_id,
            )
            if is_coro(self.callback):
                d = defer.Deferred.fromFuture(asyncio.ensure_future(self.callback(message)))
            else:
                d = threads.deferToThread(self.callback, message)
            yield d
        except Nack:
            _std_log.warning("Returning message id %s to the queue", properties.message_id)
            yield channel.basic_nack(delivery_tag=delivery_frame.delivery_tag, requeue=True)
        except Drop:
            _std_log.warning("Consumer requested message id %s be dropped", properties.message_id)
            yield channel.basic_nack(delivery_tag=delivery_frame.delivery_tag, requeue=False)
        except HaltConsumer as e:
            _std_log.info("Consumer indicated it wishes consumption to halt, shutting down")
            if e.requeue:
                yield channel.basic_nack(delivery_tag=delivery_frame.delivery_tag, requeue=True)
            else:
                yield channel.basic_ack(delivery_tag=delivery_frame.delivery_tag)
            raise e
        except Exception as e:
            _std_log.exception("Received unexpected exception from consumer %r", self)
            yield channel.basic_nack(delivery_tag=0, multiple=True, requeue=True)
            raise e
        else:
            _std_log.info(
                "Successfully consumed message from topic %s (message id %s)",
                message.topic,
                properties.message_id,
            )
            yield channel.basic_ack(delivery_tag=delivery_frame.delivery_tag)

    def _on_cancel_callback(self, frame):
        """
        Called when the consumer is canceled server-side.

        This can happen, for example, when the queue is deleted.
        To handle this, we do the necessary book-keeping to remove the consumer
        and then fire the errback on the consumer so the caller of
        :func:`fedora_messaging.api.consume` can decide what to do.

        Args:
            frame (pika.frame.Method): The cancel method from the server,
                unused here because we already know what consumer is being
                canceled.
        """
        _std_log.error("%r was canceled by the AMQP broker!", self)

        self._protocol._forget_consumer(self.queue)
        self._running = False
        self.result.errback(fail=ConsumerCanceled())

    def _read_loop_errback(self, failure):
        """
        Handle errors coming out of the read loop.

        There are two basic categories of errors: ones where the ``consumer.result``
        Deferred needs to be fired because the error is not recoverable, ones
        where we can recover from by letting the connection restart, and ones
        which are fatal for this consumer only (the queue was deleted by an
        administrator).

        Args:
            failure (twisted.python.failure.Failure): The exception raised by
                the read loop encapsulated in a Failure.
        """
        exc = failure.value
        if failure.check(pika.exceptions.ConsumerCancelled):
            # Pika 1.0.0+ raises this exception. To support previous versions
            # we register a callback (called below) ourselves with the channel.
            self._on_cancel_callback(None)
        elif failure.check(pika.exceptions.ChannelClosed):
            if exc.args[0] == 403:
                # This is a mis-configuration, the consumer can register itself,
                # but it doesn't have permissions to read from the queue,
                # so no amount of restarting will help.
                e = PermissionException(
                    obj_type="queue",
                    description=self.queue,
                    reason=failure.value.args[1],
                )
                self.result.errback(Failure(e, PermissionException))
                self.cancel()
            else:
                _std_log.exception(
                    "Consumer halted (%r) unexpectedly; " "the connection should restart.",
                    failure,
                )
        elif failure.check(error.ConnectionDone, error.ConnectionLost):
            _std_log.warning(
                "The connection to the broker was lost (%r), consumer halted; "
                "the connection should restart and consuming will resume.",
                exc,
            )
        elif failure.check(pika.exceptions.AMQPError):
            _std_log.exception(
                "An unexpected AMQP error occurred; the connection should "
                "restart, but please report this as a bug."
            )
        else:
            self.result.errback(failure)
            self.cancel()

    @defer.inlineCallbacks
    def cancel(self):
        """
        Cancel the consumer and clean up resources associated with it.
        Consumers that are canceled are allowed to finish processing any
        messages before halting.

        Returns:
            defer.Deferred: A deferred that fires when the consumer has finished
            processing any message it was in the middle of and has been successfully
            canceled.
        """
        # Remove it from protocol and factory so it doesn't restart later.
        try:
            self._protocol._forget_consumer(self.queue)
        except AttributeError:
            pass
        # Signal to the _read loop it's time to stop and wait for it to finish
        # with whatever message it might be working on, then wait for the deferred
        # to fire which indicates it is done.
        self._running = False
        yield self._read_loop
        try:
            yield self._channel.basic_cancel(consumer_tag=self._tag)
        except pika.exceptions.AMQPChannelError:
            # Consumers are tied to channels, so if this channel is dead the
            # consumer should already be canceled (and we can't get to it anyway)
            pass
        try:
            yield self._channel.close()
        except pika.exceptions.AMQPChannelError:
            pass
        if not self.result.called:
            self.result.callback(self)



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import json
from unittest.mock import AsyncMock, Mock, patch

import pika
import pytest
from twisted.internet import defer, error

from fedora_messaging.exceptions import (
    ConnectionException,
    ConsumerCanceled,
    Drop,
    HaltConsumer,
    Nack,
    PermissionException,
)
from fedora_messaging.twisted.consumer import _add_timeout, Consumer

from .utils import MockProtocol


try:
    import pytest_twisted
except ImportError:
    pytest.skip("pytest-twisted is missing, skipping tests", allow_module_level=True)


class MockDeliveryFrame:
    def __init__(self, tag, routing_key=None):
        self.delivery_tag = tag
        self.routing_key = routing_key or "routing_key"


class MockProperties:
    def __init__(self, msgid=None):
        self.headers = {}
        self.content_encoding = None
        self.message_id = msgid or "msgid"


def _queue_contents(consumer, values):
    yield from values
    consumer._running = False
    yield defer.CancelledError()


def _call_read_one(consumer, topic, headers, body):
    """Prepare arguments for the _read_one() method and call it."""
    # consumer = self.protocol._consumers["my_queue_name"]
    full_headers = {
        "fedora_messaging_schema": "base.message",
        "fedora_messaging_content_encoding": "utf-8",
    }
    full_headers.update(headers)
    queue = Mock()
    queue.get.return_value = defer.succeed(
        (
            consumer._channel,
            pika.spec.Basic.Deliver(routing_key=topic, delivery_tag="delivery_tag"),
            pika.spec.BasicProperties(headers=full_headers),
            json.dumps(body).encode("utf-8"),
        )
    )
    return consumer._read_one(queue)


def _make_consumer_with_callback(callback):
    protocol = MockProtocol(None)
    protocol._impl.is_closed = False
    consumer = Consumer("my_queue_name", callback)
    protocol._register_consumer(consumer)
    protocol.factory = Mock()
    return consumer


class TestConsumer:
    """Unit tests for the Consumer class."""

    def setup_method(self):
        self.callback = Mock()
        self.consumer = _make_consumer_with_callback(self.callback)
        self.protocol = self.consumer._protocol

    # Canceling

    @pytest_twisted.inlineCallbacks
    def test_cancel(self):
        """The cancel method must call the corresponding channel methods"""
        yield self.consumer.cancel()
        assert len(self.protocol._consumers) == 0
        # self.protocol._forget_consumer.assert_called_with("queue")
        channel = self.protocol._channel
        channel.basic_cancel.assert_called_with(consumer_tag=self.consumer._tag)
        channel.close.assert_called_with()

    @pytest_twisted.inlineCallbacks
    def test_cancel_channel_error(self):
        """Assert channel errors are caught; a closed channel cancels consumers."""
        self.consumer._channel.basic_cancel.side_effect = pika.exceptions.AMQPChannelError()
        yield self.consumer.cancel()
        assert len(self.protocol._consumers) == 0
        # consumer._protocol._forget_consumer.assert_called_with("my_queue")
        self.consumer._channel.basic_cancel.assert_called_once_with(consumer_tag=self.consumer._tag)

    # Consuming

    @pytest_twisted.inlineCallbacks
    def test_read(self, mocker):
        message = Mock(name="message")
        get_message = mocker.patch(
            "fedora_messaging.twisted.consumer.get_message", return_value=message
        )
        # Check the nominal case for _read().
        props = MockProperties()
        queue = Mock()
        queue.get.side_effect = _queue_contents(
            self.consumer,
            [
                defer.succeed(
                    (
                        self.consumer._channel,
                        MockDeliveryFrame("dt1", "rk1"),
                        props,
                        "body1",
                    )
                ),
                defer.succeed(
                    (
                        self.consumer._channel,
                        MockDeliveryFrame("dt2", "rk2"),
                        props,
                        "body2",
                    )
                ),
                defer.succeed(
                    (
                        self.consumer._channel,
                        MockDeliveryFrame("dt3", "rk3"),
                        props,
                        "body3",
                    )
                ),
            ],
        )

        yield self.consumer._read(queue)

        assert get_message.call_args_list == [
            (("rk1", props, "body1"), {}),
            (("rk2", props, "body2"), {}),
            (("rk3", props, "body3"), {}),
        ]
        assert message.queue == "my_queue_name"
        assert self.callback.call_args_list == [
            ((message,), {}),
            ((message,), {}),
            ((message,), {}),
        ]
        assert self.consumer._channel.basic_ack.call_args_list == [
            (tuple(), dict(delivery_tag="dt1")),
            (tuple(), dict(delivery_tag="dt2")),
            (tuple(), dict(delivery_tag="dt3")),
        ]

    @pytest_twisted.inlineCallbacks
    def test_read_not_running(self):
        # When not running, _read() should do nothing.
        self.consumer._running = False
        queue = Mock()
        queue.get.side_effect = lambda: defer.succeed(None)
        yield self.consumer._read(queue)
        queue.get.assert_not_called()

    @pytest_twisted.inlineCallbacks
    def test_message_invalid(self):
        # When a message is invalid, it should be Nacked.
        yield _call_read_one(self.consumer, "testing.topic", {}, "invalid-json-body")
        self.callback.assert_not_called()
        self.consumer._channel.basic_nack.assert_called_with(
            delivery_tag="delivery_tag", requeue=False
        )

    @pytest.mark.parametrize("error_class", [HaltConsumer, ValueError])
    @pytest_twisted.inlineCallbacks
    def test_read_exception(self, mocker, error_class):
        # Check that some exceptions from the callback cancel the consumer.
        self.callback.side_effect = error_class()
        message = Mock(name="message")
        mocker.patch("fedora_messaging.twisted.consumer.get_message", return_value=message)
        props = MockProperties()
        yield self.consumer._channel.queue_object.put(
            (
                self.consumer._channel,
                MockDeliveryFrame("dt1", "rk1"),
                props,
                "body1",
            )
        )
        self.consumer.cancel = Mock(return_value=defer.succeed(None))

        yield self.consumer.consume()
        yield self.consumer._read_loop
        try:
            yield self.consumer.result
        except error_class:
            pass
        else:
            pytest.fail(f"This should have raised {error_class}")

        self.consumer.cancel.assert_called_once_with()
        if error_class == HaltConsumer:
            self.consumer._channel.basic_ack.assert_called_once_with(delivery_tag="dt1")
        else:
            self.consumer._channel.basic_nack.assert_called_once_with(
                delivery_tag=0, multiple=True, requeue=True
            )

    # Handling read errors

    @pytest_twisted.inlineCallbacks
    def test_consume_channel_closed(self):
        # Check consuming when the channel is closed
        self.consumer._channel.basic_consume.side_effect = pika.exceptions.ChannelClosed(
            42, "testing"
        )
        self.consumer._read = Mock()

        try:
            yield self.consumer.consume()
        except ConnectionException:
            assert self.consumer._read_loop is None
            self.consumer._read.assert_not_called()
        else:
            pytest.fail("This should have raised ConnectionException")

    @pytest_twisted.inlineCallbacks
    def test_consume_channel_forbidden(self):
        # Check consuming when the channel is forbidden
        self.consumer._channel.basic_consume.side_effect = pika.exceptions.ChannelClosed(
            403, "testing"
        )
        self.consumer._read = Mock()

        try:
            yield self.consumer.consume()
        except PermissionException:
            assert self.consumer._read_loop is None
            self.consumer._read.assert_not_called()
        else:
            pytest.fail("This should have raised PermissionException")

    @pytest_twisted.inlineCallbacks
    def test_exit_loop_connection_done(self, mocker):
        # Check the exceptions that cause the read loop to exit.
        log = mocker.patch("fedora_messaging.twisted.consumer._std_log")
        queue = Mock()
        queue.get.side_effect = error.ConnectionDone()
        self.consumer._channel.queue_object = queue
        yield self.consumer.consume()

        yield self.consumer._read_loop
        self.callback.assert_not_called()
        log.warning.assert_called()
        assert ("The connection to the broker was lost") in log.warning.call_args[0][0]
        # Temporary error, will restart
        assert self.consumer.result.called is False

    @pytest_twisted.inlineCallbacks
    def test_exit_loop_channel_closed(self, mocker):
        # Check the exceptions that cause the read loop to exit.
        log = mocker.patch("fedora_messaging.twisted.consumer._std_log")
        queue = Mock()
        queue.get.side_effect = pika.exceptions.ChannelClosed(42, "testing")
        self.consumer._channel.queue_object = queue
        yield self.consumer.consume()

        yield self.consumer._read_loop
        self.callback.assert_not_called()
        log.exception.assert_called()
        logmsg = log.exception.call_args[0][0]
        assert "Consumer halted" in logmsg
        assert "the connection should restart" in logmsg
        # Temporary error, will restart
        assert self.consumer.result.called is False

    @pytest_twisted.inlineCallbacks
    def test_exit_loop_channel_forbidden(self):
        # Check the exceptions that cause the read loop to exit.
        queue = Mock()
        queue.get.side_effect = pika.exceptions.ChannelClosed(403, "nope!")
        self.consumer._channel.queue_object = queue
        self.consumer.consume()

        yield self.consumer._read_loop
        self.callback.assert_not_called()
        # Permanent error, no restart
        assert self.consumer._running is False
        assert self.consumer.result.called is True
        self.consumer.result.addErrback(lambda f: f.check(PermissionException))
        # The consumer should have been cancelled and the channel should have been closed
        self.consumer._channel.basic_cancel.assert_called_with(consumer_tag=self.consumer._tag)
        self.consumer._channel.close.assert_called_with()
        # Check the result's errback
        yield self.consumer.result

    @pytest_twisted.inlineCallbacks
    def test_exit_loop_consumer_cancelled(self):
        # Check the exceptions that cause the read loop to exit.
        queue = Mock()
        queue.get.side_effect = pika.exceptions.ConsumerCancelled()
        self.consumer._channel.queue_object = queue
        self.consumer.consume()

        yield self.consumer._read_loop
        self.callback.assert_not_called()
        # Permanent error, no restart
        assert self.consumer._running is False
        assert len(self.protocol._consumers) == 0
        assert self.consumer.result.called is True
        self.consumer.result.addErrback(lambda f: f.check(ConsumerCanceled))
        yield self.consumer.result

    @pytest_twisted.inlineCallbacks
    def test_exit_loop_amqp_error(self, mocker):
        # Check the exceptions that cause the read loop to exit.
        log = mocker.patch("fedora_messaging.twisted.consumer._std_log")
        queue = Mock()
        queue.get.side_effect = pika.exceptions.AMQPHeartbeatTimeout()
        self.consumer._channel.queue_object = queue
        self.consumer.consume()

        yield self.consumer._read_loop
        self.callback.assert_not_called()
        log.exception.assert_called()
        assert (
            "An unexpected AMQP error occurred; the connection should restart"
        ) in log.exception.call_args[0][0]
        # Temporary error, will restart
        assert self.consumer.result.called is False

    @pytest_twisted.inlineCallbacks
    def test_exit_loop_unknown_error(self):
        # Check the exceptions that cause the read loop to exit.
        queue = Mock()
        queue.get.side_effect = RuntimeError()
        self.consumer._channel.queue_object = queue
        self.consumer.consume()

        yield self.consumer._read_loop
        self.callback.assert_not_called()
        # Permanent error, no restart
        assert self.consumer._running is False
        assert self.consumer.result.called is True
        self.consumer.result.addErrback(lambda f: f.check(RuntimeError))
        # The consumer should have been cancelled and the channel should have been closed
        self.consumer._channel.basic_cancel.assert_called_with(consumer_tag=self.consumer._tag)
        self.consumer._channel.close.assert_called_with()
        # Check the result's errback
        yield self.consumer.result


@pytest.mark.parametrize("mock_class", [Mock, AsyncMock])
class TestConsumerCallback:
    """Unit tests for the Consumer class with a sync or async callback."""

    @pytest_twisted.inlineCallbacks
    def test_read(self, mock_class):
        # Check the nominal case for _read().
        callback = mock_class()
        consumer = _make_consumer_with_callback(callback)
        yield _call_read_one(consumer, "testing.topic", {}, {"key": "value"})
        callback.assert_called_once()
        consumer._channel.basic_ack.assert_called_once_with(delivery_tag="delivery_tag")

    @pytest_twisted.inlineCallbacks
    def test_nack(self, mock_class):
        # When the callback raises a Nack, the server should be notified.
        callback = mock_class(side_effect=Nack())
        consumer = _make_consumer_with_callback(callback)
        yield _call_read_one(consumer, "testing.topic", {}, {"key": "value"})
        callback.assert_called()
        consumer._channel.basic_nack.assert_called_with(delivery_tag="delivery_tag", requeue=True)

    @pytest_twisted.inlineCallbacks
    def test_drop(self, mock_class):
        # When the callback raises a Drop, the server should be notified.
        callback = mock_class(side_effect=Drop())
        consumer = _make_consumer_with_callback(callback)
        yield _call_read_one(consumer, "testing.topic", {}, {"key": "value"})
        callback.assert_called()
        consumer._channel.basic_nack.assert_called_with(delivery_tag="delivery_tag", requeue=False)

    @pytest.mark.parametrize("requeue", [False, True])
    @pytest_twisted.inlineCallbacks
    def test_halt(self, mock_class, requeue):
        """Assert the consumer is canceled when HaltConsumer is raised"""
        callback = mock_class(side_effect=HaltConsumer(requeue=requeue))
        consumer = _make_consumer_with_callback(callback)
        try:
            yield _call_read_one(consumer, "testing.topic", {}, {"key": "value"})
        except HaltConsumer:
            pass
        else:
            pytest.fail("This should have raised HaltConsumer")

        callback.assert_called()
        channel = consumer._channel
        if requeue:
            channel.basic_ack.assert_not_called()
            channel.basic_nack.assert_called_with(delivery_tag="delivery_tag", requeue=True)
        else:
            channel.basic_ack.assert_called_with(delivery_tag="delivery_tag")
            channel.basic_nack.assert_not_called()

    @pytest_twisted.inlineCallbacks
    def test_exception(self, mock_class):
        # On an unknown exception, the consumer should stop and all
        # unacknowledged messages should be requeued.
        callback = mock_class(side_effect=ValueError())
        consumer = _make_consumer_with_callback(callback)

        try:
            yield _call_read_one(consumer, "testing.topic", {}, {"key": "value"})
        except ValueError:
            pass
        else:
            pytest.fail("This should have raised ValueError")

        callback.assert_called()
        channel = consumer._channel
        channel.basic_nack.assert_called_with(delivery_tag=0, multiple=True, requeue=True)


class TestAddTimeout:
    """Unit tests for the _add_timeout helper function."""

    def test_twisted12_timeout(self):
        """Assert timeouts work for Twisted 12.2 (EL7)"""
        d = defer.Deferred()
        d.addTimeout = Mock(side_effect=AttributeError())
        _add_timeout(d, 0.1)

        d.addCallback(pytest.fail, "Expected errback to be called")

        def _check_failure(failure):
            assert isinstance(failure.value, defer.CancelledError)

        d.addErrback(_check_failure)

        return d

    def test_twisted12_cancel_cancel_callback(self):
        """Assert canceling the cancel call for Twisted 12.2 (EL7) works."""
        d = defer.Deferred()
        d.addTimeout = Mock(side_effect=AttributeError())
        d.cancel = Mock()
        with patch("fedora_messaging.twisted.consumer.reactor") as mock_reactor:
            _add_timeout(d, 1)
            delayed_cancel = mock_reactor.callLater.return_value
            d.callback(None)
            delayed_cancel.cancel.assert_called_once_with()



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
fedora-messaging can be configured with the
``/etc/fedora-messaging/config.toml`` file or by setting the
``FEDORA_MESSAGING_CONF`` environment variable to the path of the configuration
file.

Each configuration option has a default value.

.. contents:: Table of Configuration Options
    :local:

A complete example TOML configuration:

.. literalinclude:: ../../config.toml.example


Generic Options
===============

These options apply to both consumers and publishers.

.. _conf-amqp-url:

amqp_url
--------
The AMQP broker to connect to. This URL should be in the format described by
the :class:`pika.connection.URLParameters` documentation. This defaults to
``'amqp://?connection_attempts=3&retry_delay=5``.

.. note:: When using the Twisted consumer API, which the CLI does by default,
          any connection-related setting won't apply as Twisted manages the
          TCP/TLS connection.


.. _conf-passive-declares:

passive_declares
----------------
A boolean to specify if queues and exchanges should be declared passively (i.e
checked, but not actually created on the server). Defaults to ``False``.


.. _conf-tls:

tls
---
A dictionary of the TLS settings to use when connecting to the AMQP broker. The
default is::

    {
        'ca_cert': '/etc/pki/tls/certs/ca-bundle.crt',
        'keyfile': None,
        'certfile': None,
    }

The value of ``ca_cert`` should be the path to a bundle of CA certificates used
to validate the certificate presented by the server. The 'keyfile' and
'certfile' values should be to the client key and client certificate to use
when authenticating with the broker.

.. note:: The broker URL must use the ``amqps`` scheme. It is also possible to
          provide these setting via the ``amqp_url`` setting using a URL-encoded
          JSON object. This setting is provided as a convenient way to avoid that.


.. _conf-client-properties:

client_properties
-----------------
A dictionary that describes the client to the AMQP broker. This makes it easy
to identify the application using a connection. The dictionary can contain
arbitrary string keys and values. The default is::

    {
        'app': 'Unknown',
        'product': 'Fedora Messaging with Pika',
        'information': 'https://fedora-messaging.readthedocs.io/en/stable/',
        'version': 'fedora_messaging-<version> with pika-<version>',
    }

Apps should set the ``app`` along with any additional keys they feel will help
administrators when debugging application connections. At a minimum, the recommended
fields are:

* ``app_url``: The value of this key should be a URL to the upstream project for
  the client.
* ``app_contacts_email``: One or more emails of maintainers to contact with
  questions (if, for example, a client is misbehaving, or a service disruption
  is about to occur).

Do not use the ``product``, ``information``, and ``version`` keys as these will
be set automatically.


.. _conf-exchanges:

exchanges
---------
A dictionary of exchanges that should be present in the broker. Each key should
be an exchange name, and the value should be a dictionary with the exchange's
configuration.  Options are:

* ``type`` - the type of exchange to create.

* ``durable`` - whether or not the exchange should survive a broker restart.

* ``auto_delete`` - whether or not the exchange should be deleted once no queues
  are bound to it.

* ``arguments`` - dictionary of arbitrary keyword arguments for the exchange,
  which depends on the broker in use and its extensions.

For example::

    {
        'my_exchange': {
            'type': 'fanout',
            'durable': True,
            'auto_delete': False,
            'arguments': {},
        },
    }

The default is to ensure the 'amq.topic' topic exchange exists which should be
sufficient for most use cases.

.. _conf-log-config:

log_config
----------
A dictionary describing the logging configuration to use, in a format accepted
by :func:`logging.config.dictConfig`.

.. note:: Logging is only configured for consumers, not for producers.


Publisher Options
=================

The following configuration options are publisher-related.

.. _conf-publish-exchange:

publish_exchange
----------------
A string that identifies the exchange to publish to. The default is
``amq.topic``.


.. _conf-topic-prefix:

topic_prefix
------------
A string that will be prepended to topics on sent messages.
This is useful to migrate from fedmsg, but should not be used otherwise.
The default is an empty string.

.. _conf-publish-priority:

publish_priority
----------------
A number that will be set as the priority for the messages. The range of
possible priorities depends on the ``x-max-priority`` argument of the
destination queue, as described in `RabbitMQ's priority documentation`_.
The default is ``None``, which RabbitMQ will interpret as zero.

.. _RabbitMQ's priority documentation: https://www.rabbitmq.com/priority.html

.. _sub-config:

Consumer Options
================

The following configuration options are consumer-related.

.. _conf-queues:

queues
------
A dictionary of queues that should be present in the broker. Each key should be
a queue name, and the value should be a dictionary with the queue's configuration.
Options are:

* ``durable`` - whether or not the queue should survive a broker restart. This is
  set to ``False`` for the default queue.

* ``auto_delete`` - whether or not the queue should be deleted once the
  consumer disconnects. This is set to ``True`` for the default queue.

* ``exclusive`` - whether or not the queue is exclusive to the current
  connection. This is set to ``False`` for the default queue.

* ``arguments`` - dictionary of arbitrary keyword arguments for the queue, which
  depends on the broker in use and its extensions. This is set to ``{}`` for the
  default queue

For example::

    {
        'my_queue': {
            'durable': True,
            'auto_delete': True,
            'exclusive': False,
            'arguments': {},
        },
    }


.. _conf-bindings:

bindings
--------
A list of dictionaries that define queue bindings to exchanges that consumers
will subscribe to. The ``queue`` key is the queue's name. The ``exchange`` key
should be the exchange name and the ``routing_keys`` key should be a list of
routing keys. For example::

    [
        {
            'queue': 'my_queue',
            'exchange': 'amq.topic',
            'routing_keys': ['topic1', 'topic2.#'],
        },
    ]

This would create two bindings for the ``my_queue`` queue, both to the
``amq.topic`` exchange. Consumers will consume from both queues.

.. _conf-callback:

callback
--------
The Python path of the callback. This should be in the format
``<module>:<object>``. For example, if the callback was called "my_callback"
and was located in the "my_module" module of the "my_package" package, the path
would be defined as ``my_package.my_module:my_callback``. The default is None.

Consult the :ref:`consumers` documentation for details on implementing a
callback.

.. _conf-consumer-config:

consumer_config
---------------
A dictionary for the consumer to use as configuration. The consumer should
access this key in its callback for any configuration it needs. Defaults to
an empty dictionary. If, for example, this dictionary contains the
``print_messages`` key, the callback can access this configuration with::

    from fedora_messaging import config

    def callback(message):
        if config.conf["consumer_config"]["print_messages"]:
            print(message)


.. _conf-qos:

qos
---
The quality of service settings to use for consumers. This setting is a
dictionary with two keys. ``prefetch_count`` specifies the number of messages
to pre-fetch from the server. Pre-fetching messages improves performance by
reducing the amount of back-and-forth between client and server. The downside
is if the consumer encounters an unexpected problem, messages won't be returned
to the queue and sent to a different consumer until the consumer times out.
``prefetch_size`` limits the size of pre-fetched messages (in bytes), with 0
meaning there is no limit. The default settings are::

    {
        'prefetch_count': 10,
        'prefetch_size': 0,
    }
"""


import copy
import logging
import logging.config
import os
from importlib.metadata import version


try:
    import tomllib
except ImportError:
    import tomli as tomllib

from . import exceptions


_log = logging.getLogger(__name__)

_fedora_version = version("fedora_messaging")
_pika_version = version("pika")

# By default, use a server-generated queue name
_default_queue_name = ""

#: The default configuration settings for fedora-messaging. This should not be
#: modified and should be copied with :func:`copy.deepcopy`.
DEFAULTS = dict(
    amqp_url="amqp://?connection_attempts=3&retry_delay=5",
    #: The default client properties reported to the AMQP broker in the "start-ok"
    #: method of the connection negotiation. This allows the broker administrators
    #: to easily identify what a connection is being used for and the client's
    #: capabilities.
    client_properties={
        "app": "Unknown",
        "product": "Fedora Messaging with Pika",
        "information": "https://fedora-messaging.readthedocs.io/en/stable/",
        "version": f"fedora_messaging-{_fedora_version} with pika-{_pika_version}",
    },
    publish_exchange="amq.topic",
    topic_prefix="",
    publish_priority=None,
    passive_declares=False,
    exchanges={
        "amq.topic": {
            "type": "topic",
            "durable": True,
            "auto_delete": False,
            "arguments": {},
        }
    },
    queues={
        _default_queue_name: {
            "durable": False,
            "auto_delete": True,
            "exclusive": True,
            "arguments": {},
        }
    },
    bindings=[{"queue": _default_queue_name, "exchange": "amq.topic", "routing_keys": ["#"]}],
    qos={"prefetch_size": 0, "prefetch_count": 10},
    callback=None,
    consumer_config={},
    tls={"ca_cert": None, "certfile": None, "keyfile": None},
    log_config={
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {"simple": {"format": "[%(name)s %(levelname)s] %(message)s"}},
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "formatter": "simple",
                "stream": "ext://sys.stdout",
            }
        },
        "loggers": {
            "fedora_messaging": {
                "level": "INFO",
                "propagate": False,
                "handlers": ["console"],
            }
        },
        # The root logger configuration; this is a catch-all configuration
        # that applies to all log messages not handled by a different logger
        "root": {"level": "WARNING", "handlers": ["console"]},
    },
)


def validate_bindings(bindings):
    """
    Validate the bindings configuration.

    Raises:
        exceptions.ConfigurationException: If the configuration provided is of an
            invalid format.
    """
    if not isinstance(bindings, (list, tuple)):
        raise exceptions.ConfigurationException(
            f"bindings must be a list or tuple of dictionaries, but was a {type(bindings)}"
        )

    for binding in bindings:
        missing_keys = []
        for key in ("exchange", "routing_keys"):
            if key not in binding:
                missing_keys.append(key)
        if missing_keys:
            raise exceptions.ConfigurationException(
                "a binding is missing the following keys from its settings "
                f"value: {missing_keys}"
            )

        if not isinstance(binding["routing_keys"], (list, tuple)):
            raise exceptions.ConfigurationException(
                "routing_keys must be a list or tuple, but was a {}".format(
                    type(binding["routing_keys"])
                )
            )


def validate_queues(queues):
    """
    Validate the queues configuration.

    Raises:
        exceptions.ConfigurationException: If the configuration provided is of an
            invalid format.
    """
    if not isinstance(queues, dict):
        raise exceptions.ConfigurationException(
            "'queues' must be a dictionary mapping queue names to settings."
        )

    for queue, settings in queues.items():
        if not isinstance(settings, dict):
            raise exceptions.ConfigurationException(
                f"the {queue} queue in the 'queues' setting has a value of type {type(settings)}, "
                "but it should be a dictionary of settings."
            )
        missing_keys = []
        for key in ("durable", "auto_delete", "exclusive", "arguments"):
            if key not in settings:
                missing_keys.append(key)
        if missing_keys:
            raise exceptions.ConfigurationException(
                f"the {queue} queue is missing the following keys from its settings "
                f"value: {missing_keys}"
            )


def validate_client_properties(props):
    """
    Validate the client properties setting.

    This will add the "version", "information", and "product" keys if they are
    missing. All other keys are application-specific.

    Raises:
        exceptions.ConfigurationException: If any of the basic keys are overridden.
    """
    for key in ("version", "information", "product"):
        # Nested dictionaries are not merged so key can be missing
        if key not in props:
            props[key] = DEFAULTS["client_properties"][key]
        # Don't let users override these as they identify this library in AMQP
        if props[key] != DEFAULTS["client_properties"][key]:
            raise exceptions.ConfigurationException(
                f'"{key}" is a reserved keyword in client_properties'
            )


class LazyConfig(dict):
    """This class lazy-loads the configuration file."""

    loaded = False

    def __getitem__(self, *args, **kw):
        if not self.loaded:
            self.load_config()
        return super().__getitem__(*args, **kw)

    def get(self, *args, **kw):
        if not self.loaded:
            self.load_config()
        return super().get(*args, **kw)

    def pop(self, *args, **kw):
        raise exceptions.ConfigurationException("Configuration keys cannot be removed!")

    def copy(self, *args, **kw):
        if not self.loaded:
            self.load_config()
        return super().copy(*args, **kw)

    def update(self, *args, **kw):
        if not self.loaded:
            self.load_config()
        return super().update(*args, **kw)

    def setup_logging(self):
        if not self.loaded:
            self.load_config()
        logging.config.dictConfig(self["log_config"])

    def _validate(self):
        """
        Perform checks on the configuration to assert its validity

        Raises:
            ConfigurationException: If the configuration is invalid.
        """
        for key in self:
            if key not in DEFAULTS:
                raise exceptions.ConfigurationException(
                    f"Unknown configuration key {key!r}! Valid configuration keys are"
                    f" {list(DEFAULTS.keys())}"
                )

        validate_queues(self["queues"])
        validate_bindings(self["bindings"])
        validate_client_properties(self["client_properties"])

    def load_config(self, config_path=None):
        """
        Load application configuration from a file and merge it with the default
        configuration.

        If the ``FEDORA_MESSAGING_CONF`` environment variable is set to a
        filesystem path, the configuration will be loaded from that location.
        Otherwise, the path defaults to ``/etc/fedora-messaging/config.toml``.
        """
        self.loaded = True
        config = copy.deepcopy(DEFAULTS)

        if config_path is None:
            if "FEDORA_MESSAGING_CONF" in os.environ:
                config_path = os.environ["FEDORA_MESSAGING_CONF"]
            else:
                config_path = "/etc/fedora-messaging/config.toml"

        if os.path.exists(config_path):
            _log.info(f"Loading configuration from {config_path}")
            with open(config_path, "rb") as fd:
                try:
                    file_config = tomllib.load(fd)
                    for key in file_config:
                        config[key.lower()] = file_config[key]
                except tomllib.TOMLDecodeError as e:
                    msg = f"Failed to parse {config_path}: {e}"
                    raise exceptions.ConfigurationException(msg) from e
        else:
            _log.info(f"The configuration file, {config_path}, does not exist.")

        self.update(config)
        self._validate()
        return self


#: The configuration dictionary used by fedora-messaging and consumers.
conf = LazyConfig()
